//  This file handles alpha binding---the initial binding we do after an
//  expression is read and expanded but before it is evaluated.  We simply
//  look up each |id| (|EXPR_ID|) and turn it into either a global pointer
//  (|EXPR_GLOBAL|) or a scoped pointer (|EXPR_SCOPED| or |EXPR_LOCAL|).
//  It destructs its argument.
//
//  The following special forms are treated differently:
//
//  |quote| is not examined, just returned literally.
//
//  |quasiquote| and |unquote| increase and decrease the qqlev, saying
//     whether we look at things or not.
//
//  |lambda| adds its argument list to a new scope built off the current
//     scope and recurses.
//
//  |let| adds each new variable to the new scope, evaluates their
//     initializers in the previous scope and evaluates the body in the
//     new scope.
//
//  |define| is converted to a lambda expression if necessary.
//  |defmacro| is converted into a lambda expression if necessary in the
//     same way as |define|.

#ifdef AMIGA
#define ScopeEntry z // Hack for Amiga compiler
#endif

#include "Scope.h"

//  If we don't find an |id| in our scope, we `initialize' it in the global
//  scope to Null.
//
//  Note that |maxuse| is (in Ada-speak) an {\bf inout} variable; the storage
//  it points to must be initialized to zero.  This is so the maximum can be
//  accumulated over several calls, such as to arguments or in a |begin|
//  sequence.
//
//  We need to pass a |maxuse| for any expression that might be actually
//  evaluated, so that we know how large a local environment vector to
//  allocate.

Expr *Expr::alphaBind(Scope *scop, int *maxuse, int nv =1,
                                                int clev =1, int qqlev =0) {
   ScopeEntry *se = 0 ;
   Expr *ret = Null ;
   extern Scope *scope ;

//  What?  You're not used to seeing the |default| case at the top?
//  Well, get used to it; I do it a lot.

   switch(type) {
default:
      return this ;
case EXPR_ID:
      if (qqlev > 0)
         return this ;
      se = scop->findInTree(val.id) ;
      if (se == 0) {
         scope->add(val.id, Undefined) ;
         se = scop->findInTree(val.id) ;
         if (se == NULL)
            error("! internal error; added id didn't take") ;
      }
      if (!se->exprp()) {
         error("invalid use of a macro") ;
         this->remove() ;
         return Null ;
      }

//  We need to differentiate a global from a scoped variable from a local.
//  This is easy; a scoped variable, at this point, will return |EXPR_SCOPED|;
//  a local variable will return the same, but the index will be greater than
//  or equal to |clev|; a global will return something else.  (Like a value.)

      ret = se->value() ;
      if (! ret->boundp()) {
         ret->remove() ;
         ret = new Expr(se) ;
      } else if (ret->scopeIndex() >= clev) {
         ret->remove() ;
         ret = new Expr(se, ret->scopeIndex() - clev, EXPR_LOCAL) ;
      }
      this->remove() ;
      return ret ;

//  The complex case checks the first component of the list.  If it is an
//  |id|, and that |id| is one of our special cases, we treat it ourselves.
//  Otherwise, we simply recurse, consing up a new list as we go.
//  If we do a |quote| or a |defmacro|, we only bind the name; everything
//  else stays as is.

case EXPR_COMPLEX:
      const char *id = this->carid() ;
      if (id) {
         if (strcmp(id, "quote")==0) {
            ret = new Expr(this->car()->copy()->
             alphaBind(scop, maxuse, nv, clev, qqlev), this->cdr()->copy()) ;
            this->remove() ;
            return ret ;
         }
         if (strcmp(id, "unquote")==0)
            qqlev-- ;
         else if (strcmp(id, "quasiquote")==0) {
            ret = new Expr(this->car()->copy()->
                                   alphaBind(scop, maxuse, nv, clev, qqlev),
                          this->cdr()->copy()->
                             alphaBind(scop, maxuse, nv, clev, qqlev + 1)) ;
            this->remove() ;
            return ret ;
         }
         if (qqlev > 0) {
            id = 0 ;

//  For a lambda expression, the first argument must be a list of identifiers.
//  We create a new scope and add each of these identifiers to it, giving them
//  the value from |nv|, incrementing |nv| each time.  (|nv| will be used as
//  an index into an array to quickly find the values from our bound
//  identifiers.)  Then, we expand the body from this new scope and |nv|.
//
//  Our final expression looks like |(lambda (a b c d . e) . body)|, where
//  `|a|' is the lexical level, `|b|' is the arity, `|c|' is the maximum
//  number of enclosed local variables (so we can allocate a scope array),
//  `|d|' is our hash table so we can delete it with the lambda, and `|e|'
//  is a list of our arguments, so we can access their types.
//
//  The maximum number of enclosed local variables
//  has to be set during later expansions, so we pass a pointer to the value
//  to set.

         } else if (strcmp(id, "lambda")==0) {
            Scope *newscope = new Scope(scop) ;
            Expr *formals = this->cdr()->car() ;
            Expr *t = new Expr(new Expr(newscope), Null) ;
            ret = new Expr(nv) ;
            Expr *tt = t ;
            int onv = nv ;
            while (formals != Null) {
               if (formals->consp()) {
                  id = newscope->declare(formals->car(), nv) ;
                  tt->setcdr(new Expr(
                  new Expr(newscope->find(id), nv++ - onv, EXPR_LOCAL), Null)) ;
                  tt = tt->cdr() ;
                  formals = formals->cdr() ;
               } else {
                  id = newscope->declare(formals, nv) ;
                  tt->setcdr(
                       new Expr(newscope->find(id), nv++ - onv, EXPR_LOCAL)) ;
                  formals = Null ;
               }
            }
            ret = new Expr(ret,
                     new Expr(new Expr(nv-ret->numv()),
                        new Expr(new Expr(nv), t))) ;
            maxuse = (int *)(ret->cdr()->cdr()->car()->valAddress()) ;

//  If there is not exactly one expression, we wrap a |begin| around the
//  statement part.

            if (this->length() != 3) {
               ret = new Expr(ret,
                        new Expr((new Expr(Literal_begin,
                                          this->cdr()->cdr()->copy()))->
                          alphaBind(newscope, maxuse, nv, onv), Null)) ;
            } else {
               ret = new Expr(ret, this->cdr()->cdr()->copy()->
                                 alphaBind(newscope, maxuse, nv, onv)) ;
            }
            ret = new Expr(this->car()->copy()->
                               alphaBind(scop, maxuse, nv, clev), ret) ;

//  For a |let| expression, we do something very similar, but just a bit more
//  complex.  (We could just turn |let| into a |lambda| expression, but the
//  final result is simpler this way.)
//  Anyway, the first argument of the |let| gets turned into
//  a 3 element list, where the first element is the current level, the
//  second is the arity, and the third is the list of parameters.
//
//  Note that the difference between |let| and |let*| is simply
//  whether the original scope or the accumulated scope,
//  respectively, is used when evaluating the initialization.

         } else if (strcmp(id, "let")==0 || strcmp(id, "let*")==0) {
            Scope *newscope = new Scope(scop) ;
            Expr *nthis ;
            int nclev ;

//  If this is a named let, we treat it just slightly differently.
//  A named let is treated more like a closure.

            Scope *tscope = (id[3] == '*') ? newscope : scop ;
            int namedlet = this->cdr()->car()->idp() ;
            if (namedlet) {
               if (tscope == newscope) {
                  error("named let*'s not supported; treating as let") ;
                  tscope = scop ;
               }
               ret = new Expr(nv) ;
               nthis = this->cdr() ;
               id = nthis->carid() ;
               newscope->addBound(id, nv++) ;
               nclev = nv ;
            } else {
               nthis = this ;
               nclev = clev ;
            }
            Expr *formals = nthis->cdr()->car() ;
            Expr *t, *tt ;
            int *nmaxuse = maxuse ;
            tt = new Expr(new Expr(newscope), Null) ;
            t = new Expr(tt, Null) ;
            if (namedlet) {
               ret = new Expr(ret, new Expr(new Expr(formals->length()),
                               new Expr(new Expr(nv + formals->length()), t))) ;
               nmaxuse = (int *)(ret->cdr()->cdr()->car()->valAddress()) ;
            } else {
               ret = t ;
            }
            Expr *rep ;
            while (formals != Null) {
               rep = new Expr(formals->car()->cdr()->car()->copy()->
                           alphaBind(tscope, nmaxuse, nv, clev), Null) ;
               id = newscope->declare(formals->car()->car(), nv) ;
               tt->setcdr(new Expr(
                new Expr(newscope->find(id), nv++ - clev, EXPR_LOCAL), Null)) ;
               tt = tt->cdr() ;
               t->setcdr(rep) ;
               t = t->cdr() ;
               formals = formals->cdr() ;
            }
            if (nv > *maxuse)
               *maxuse = nv ;

//  Once again, if we don't have exactly one expression, we wrap a begin
//  around things.  In any case, each of the three |let| forms use the
//  same `evaluator'; all of the difference is in the binding.

            if (nthis->length() != 3) {
               ret = new Expr(ret,
                        new Expr((new Expr(Literal_begin,
                                          nthis->cdr()->cdr()->copy()))->
                       alphaBind(newscope, nmaxuse, nv, nclev), Null)) ;
            } else {
               ret = new Expr(ret, nthis->cdr()->cdr()->copy()->
                              alphaBind(newscope, nmaxuse, nv, nclev)) ;
            }
            ret = new Expr((new Expr((namedlet ? "namedlet" : "let")))->
                             alphaBind(scop, nmaxuse, nv, nclev), ret) ;
            if (*nmaxuse > *maxuse)
               *maxuse = *nmaxuse ;

//  For a |define|, we simply check the second argument.  If it's a cons
//  list, then we transform the expression into a |lambda|, and continue.
//  (There are some pretty hairy machinations below to handle types.)
//  Otherwise, we treat most of it as any other expression, except for
//  the second argument, which we handle inline here, since it might be
//  typed.
//
//  We treat |defmacro| in exactly the same way.

         } else if (strcmp(id, "define")==0 || strcmp(id, "defmacro")==0) {
            ret = this->cdr()->car() ;
            if (ret->consp() && strcmp(ret->carid(), "type") != 0) {
               Type *t = new Type(ret->car()) ;
               id = t->typeId() ;
               if (id == 0) {
                  error("bad define form") ;
                  id = (char *)1 ;
                  ret = Null ;
               } else {
                  ret = new Expr(Literal_lambda,
                           new Expr(this->cdr()->car()->cdr()->copy(),
                                    this->cdr()->cdr()->copy())) ;
                  ret = new Expr(this->car()->copy(),
                           new Expr(new Expr(Literal_type,
                   new Expr(new Expr(id),
                      new Expr(this->cdr()->car()->copy(), Null))),
                              new Expr(ret, Null))) ;
                  ret = ret->alphaBind(scop, maxuse, nv) ;
               }
               delete t ;

//  Define is only valid at the outermost level, to define globals.
//  Locals are declared in |let| or |lambda| and |set!|.  If we want
//  to allow nested |define|s, at some point, this is where we would
//  do it.
//
//  For now we make sure what we are adding is a global, and complain
//  if it's not.  In either case, we make it a global.

            } else {
               id = scope->declare(ret) ;
               if (id == 0) {
                  id = (char *)1 ;
                  se = 0 ;
     error("expected identifier in second position of define or defmacro") ;
               } else {
                  se = scop->findInTree(id) ;
                  if (se->valueNoCopy()->boundp()) {
                     error2("can only define global types; %s is local", id) ;
                     se = scope->find(id) ;
                  }
               }
               ret = new Expr(this->car()->copy()->alphaBind(scop, maxuse, nv),
                        new Expr((se ? (new Expr(se)) : Null),
                           this->cdr()->cdr()->copy()->
                                              alphaBind(scop, maxuse, nv))) ;
            }
         } else
            id = 0 ;
      }

//  If we get here, then we've assumed we're treating things as an ordinary
//  list, which isn't particularly difficult.

      if (id == 0 && this != Null) {
         ret = new Expr(this->car()->copy()->
                                 alphaBind(scop, maxuse, nv, clev, qqlev),
                        this->cdr()->copy()->
                                 alphaBind(scop, maxuse, nv, clev, qqlev)) ;
      }
      this->remove() ;
      return ret ;
   }
}
