//  The utility routines we need to maintain |EvalState|.

#include "EvalState.h"

#define STARTSIZE (100)
#define FUZZ (20)

//  This function creates a new evaluation state from an expression,
//  assuming none (global) context.

EvalState::EvalState(Expr *e, int locals, List *rQ) {
   int locs, cont ;
   estack = new Expr *[STARTSIZE] ;
   esp = estack ;
   esend = estack + STARTSIZE - FUZZ ;
   essize = STARTSIZE ;
   if (locals) {
      *esp++ = (Expr *)RETURN_ME ; // cont
      for (locs=locals; locs; locs--)
         *esp++ = Null ;
      *esp++ = (Expr *)locals ;
      cont = POP_LOCALS ;
   } else
      cont = RETURN_ME ;
   pushState(0, estack + 1, e, 0, cont) ;
   waitcount = 0 ;
   readyQ = rQ ;
   eslink.setLink((elem)this) ;
}

//  This function clones an evaluation state, set up for execution of
//  another process.  This is necessary when we have to `spawn' another
//  process, since the two processes must share environments.

EvalState::EvalState(EvalState *es, Expr *e, int cont) {
   int locals ;
   if (! e->lambdap() || e->lambdaArity() != 0) {
      error("Expected a thunk") ;
      e = Null ;
      locals = 0 ;
   } else {
      locals = e->lambdaMax() - e->lambdaLow() ;
   }
   int locs ;
   estack = new Expr *[STARTSIZE] ;
   esp = estack ;
   esend = estack + STARTSIZE - FUZZ ;
   essize = STARTSIZE ;
   *esp++ = (Expr *)es ;
   if (locals) {
      *esp++ = (Expr *)cont ; // cont
      for (locs=locals; locs; locs--)
         *esp++ = Null ;
      *esp++ = (Expr *)locals ;
      cont = POP_LOCALS ;
   }
   pushState(0, estack + 2, new Expr(e, Null), 0, cont) ;
   waitcount = 0 ;
   readyQ = es->readyQ ;
   eslink.setLink((elem)this) ;
}

EvalState::~EvalState(VOID) {
   delete estack ;
}

//  This function pushes the current execution context from local
//  variables.  It is useful for stopping the computation so it can
//  be later restarted.

void EvalState::pushState(Expr **scopeenv, Expr **localenv, Expr *exp,
                                            int funklock, int cont)  {
#ifdef NOTYET
   if (localenv < getBase())
      error("! pushing a bad local environment") ;
#endif
   pushInt(localenv - getBase()) ; pushExprPP(scopeenv) ;
   push(exp) ; pushInt(funklock) ; pushInt(cont) ;
}

//  This is the analogous function that works the other way.

void EvalState::popState(Expr ***scopeenv, Expr ***localenv, Expr **exp,
                                           int *funklock, int *cont) {
   *cont = popInt() ; *funklock = popInt() ; *exp = pop() ;
   *scopeenv = popExprPP() ; *localenv = getBase() + popInt() ;
#ifdef NOTYET
   if (*localenv < getBase())
      error("! popped a bad local environment") ;
#endif
}

//  This function marks the first time in my life I have used a pointer to
//  a pointer to a pointer.  (I wrote the following function before I
//  wrote the immediately preceding function.)  The reason is that we need
//  to modify the
//  |localenv| pointer if we go recursive and grow the stack too large.
//  Since |localenv| is a pointer to an array of pointers to bindings,
//  we need three of those stars.

void EvalState::growStack(Expr ***locptr) {
   int nessize = 2 * essize ;
   Expr **nstack = new Expr *[nessize] ;
   int depth = esp - estack ;
   bcopy(estack, nstack, essize * sizeof(Expr **)) ;
   *locptr = nstack + (*locptr - estack) ;
   estack = nstack ;
   esp = estack + depth ;
   essize = nessize ;
   esend = estack + essize - FUZZ ;
   if (debugging('s'))
      printf("Growing the stack to %d\n", essize) ;
}

//  Schedules an evaluation.  Just adds to the current list.

void EvalState::schedule(void) {
   readyQ->append(&eslink) ;
}

//  Unschedules an evaluation.  Just pops from the current list.

void EvalState::unschedule(void) {
   readyQ->removeNoDealloc(&eslink) ;
}

//  We need a function to initialize our forms in our global symbol table.
//  This is also a convenient list of our forms.  The primitives are in
//  |Prims.cc|.
//
//  And since we are in an initialization context, we initialize our
//  global dictionary.

void initforms(Scope *scope) {
   scope->add("if", new Expr(IF_FORM, EXPR_FORM)) ;
   scope->add("begin", new Expr(BEGIN_FORM, EXPR_FORM)) ;
   scope->add("model", new Expr(MODEL_FORM, EXPR_FORM)) ;
   scope->add("sim", new Expr(SIM_FORM, EXPR_FORM)) ;
   scope->add("stim", new Expr(STIM_FORM, EXPR_FORM)) ;
   scope->add("construct", new Expr(CONS_FORM, EXPR_FORM)) ;
   scope->add("define", new Expr(DEFINE_FORM, EXPR_FORM)) ;
   scope->add("set!", new Expr(SET_FORM, EXPR_FORM)) ;
   scope->add("lambda", new Expr(LAMBDA_FORM, EXPR_FORM)) ;
   scope->add("let", new Expr(LET_FORM, EXPR_FORM)) ;
   scope->add("namedlet", new Expr(NAMED_LET_FORM, EXPR_FORM)) ;
   scope->add("time", new Expr(TIME_FORM, EXPR_FORM)) ;
   scope->add("display-type", new Expr(DISPLAY_TYPE_FORM, EXPR_FORM)) ;
   scope->add("defmacro", new Expr(DEF_MAC_FORM, EXPR_FORM)) ;
   scope->add("par", new Expr(SPAWN_FORM, EXPR_EV_FORM)) ;
   scope->add("quasiquote", new Expr(QQUOTE_FORM, EXPR_FORM)) ;
   scope->add("unquote", new Expr(UNQUOTE_FORM, EXPR_FORM)) ;
   scope->add("unquote-splicing", new Expr(UNQUOTES_FORM, EXPR_FORM)) ;
   scope->add("quote", new Expr(QUOTE_FORM, EXPR_FORM)) ;
   scope->add("in", new Expr(IN_FORM, EXPR_EV_FORM)) ;
   scope->add("out", new Expr(OUT_FORM, EXPR_EV_FORM)) ;
   scope->add("makegraph", new Expr(MAKE_GRAPH_FORM, EXPR_EV_FORM)) ;
}
