//  This file handles macro expansion of an expression.  The only complex
//  part is the |complex| part.  We delete the expression as we create
//  the new expression, so this also functions as a delete.

#include "Scope.h"

//  We need access to the external scope to prevent infinite recursion.

extern Scope *scope ;

//  The maximum depth to which we expand macros is defined here; this does
//  not prevent the user from defining recursive macros, but it keeps us
//  from blowing up the stack if he tries to use them.

#define MAXMACROEXPAND (60)

Expr *Expr::macroexpand(int level =0) {
   ScopeEntry *se = 0 ;
   if (level > MAXMACROEXPAND) {
      error("macro expansion past maximum number of levels; returning Null") ;
      this->remove() ;
      return Null ;
   }
   switch(type) {
case EXPR_ID:
      se = ::scope->findInTree(val.id) ;
      if (se && se->simpleMacrop()) {
         if (debugging('m')) {
            printf("Expanding ") ;
            this->display(stdout) ;
            printf("\n") ;
         }
         this->remove() ;
         return se->value()->macroexpand(level+1) ;
      } else
         return this ;

//  The complex case checks the first component of the list.  If it is an
//  |id|, and that |id| is a complex macro, we cons up an expression to
//  be evaluated, passing in the appropriate closure.

case EXPR_COMPLEX:
      const char *id = this->carid() ;
      Expr *ret = Null ;
      if (id && (se = ::scope->findInTree(id)) && se->complexMacrop()) {
         Expr *val = se->value() ;
         if (debugging('m')) {
            printf("Expanding ") ;
            this->display(stdout) ;
            printf("\nBy ") ;
            val->display(stdout) ;
            printf("\n") ;
         }
         int i = (val->lambdap() ? val->lambdaMax() : 20) ;
         val = new Expr(val, this->cdr()->quotifyList()) ;
         val = val->eval(i) ;
         if (debugging('m')) {
            printf("Expanded to ") ;
            val->display(stdout) ;
            printf("\n") ;
         }
         ret = val->macroexpand(level+1) ;
         this->remove() ;

//  We need to turn off macro expansions when we are reading a macro.
//  This is somewhat of a kludge, unfortunately.

      } else if ((id && strcmp(id, "defmacro")==0) || this == Null) {
         ret = this ;
      } else {

//  Otherwise it's just an ordinary list.

         ret = new Expr(this->car()->copy()->macroexpand(level),
                        this->cdr()->copy()->macroexpand(level)) ;
         this->remove() ;
      }
      return ret ;
default:
      return this ;
   }
}

//  This function takes a list and returns a list of quoted args for
//  our macro expander.

Expr *Expr::quotifyList(VOID) {
   if (this == Null)
      return Null ;
   else {
      Expr *head = this->car() ;
      int dummy ;
      if (head->consp() || head->idp())
         head = new Expr(Literal_quote->alphaBind(::scope, &dummy),
                         new Expr(head->copy(), Null)) ;
      else
         head = head->copy() ;
      return new Expr(head, this->cdr()->quotifyList()) ;
   }
}
