//  The implementation of the |Expr| class is taken from the world of Lisp.

#include "HWSG.h"
#include "Key.h"
#include "display.h"

Expr *Expr::car() {
   if (type == EXPR_COMPLEX)
      return (val.cons.car) ;
   else
      return (Null) ;
}

Expr *Expr::cdr() {
   if (type == EXPR_COMPLEX)
      return (val.cons.cdr) ;
   else
      return (Null) ;
}

//  The |charid| function is somewhat unique; it makes sure that the
//  current |Expr| is a list and that the car is an identifier, and
//  then it returns a pointer to that identifier.  Useful for evaluating,
//  since we need to look up the identifier before we know how to
//  evaluate things.

const char *Expr::carid() {
   if (type == EXPR_COMPLEX && val.cons.car->type == EXPR_ID)
      return(val.cons.car->val.id) ;
   else
      return(0) ;
}

//  If we blow away an |Expr|, we need to make sure to blow away anything
//  that might be attached.  We need to make sure we never delete |Null|,
//  though.  If we try to delete |Null|, we |forever()| it instead.
//
//  Under C++ 1.2, we've got to make sure we increment the object deleted count,
//  because we're bypassing the |Object| destructor, if the object is not
//  |Null|.  Under C++ 2.0, we've got to decrement the deleted count if the
//  object {\it is\/} |Null|, since the |Object| destructor is called.
//
//  All of this just to get reference counts to work correctly!
//
//  There is definitely a bug in this code---any recursive arrays or lists
//  are memory leaks.  Things `work' but eventually VM will be exhausted.
//  This is not true for lambda expressions, since these may be legitimately
//  recursive.

#ifdef CPLUSV2POINT0
void Expr::operator delete(void *op) {
   Expr *t = (Expr *)op ;
#define PRE t->
#define THIS t
#else
Expr::~Expr() {
#define PRE
#define THIS this
#endif

   if (THIS == Null) {
      PRE forever() ;
#ifdef CPLUSV2POINT0
      Object_deleted-- ;
#endif
   } else {
#ifndef CPLUSV2POINT0
      Object_deleted++ ;
#endif
#ifdef DEBUG
      printf("Deleting ") ; PRE display(stdout) ; printf("\n") ;
#endif
      switch(PRE type) {
case EXPR_COMPLEX:
         PRE refcount = 1000000 ; // for recursive references
         PRE val.cons.cdr->remove() ;
         PRE val.cons.car->remove() ;
         break ;
case EXPR_LAMBDA:
         PRE refcount = 1000000 ; // for recursive reference
         while (PRE val.lambda.low > 0) {
            PRE val.lambda.low-- ;
            PRE val.lambda.SA[PRE val.lambda.low]->remove() ;
         }
         (*(--(PRE val.lambda.SA)))->remove() ; // delete hash type
         (*(--(PRE val.lambda.SA)))->remove() ; // delete arg list
         delete PRE val.lambda.SA ;
         break ;
case EXPR_ARRAY:
         PRE refcount = 1000000 ; // for recursive references
         while (PRE val.array.len > 0) {
            PRE val.array.len-- ;
            PRE val.array.arr[PRE val.array.len]->remove() ;
         }
         delete PRE val.array.arr ;
         break ;
case EXPR_BIND:
         PRE refcount = 8 ; // to prevent bad recursive deletions
         PRE val.cons.car->remove() ;
         break ;
case EXPR_ID:
case EXPR_STRING:
         delete PRE val.id ;
         break ;
case EXPR_SCOPE:
         delete PRE val.scope ;
         break ;
case EXPR_FILE:
         if (PRE val.file) {
            fclose(PRE val.file) ;
            PRE val.file = 0 ;
         }
         break ;
case EXPR_GRAPH:
         if (PRE val.graph) {
            delete PRE val.graph ;
            PRE val.graph = 0 ;
         }
         break ;
case EXPR_WINDOW:
         if (PRE val.w.graph) {
            PRE val.w.graph->remove() ;
            PRE val.w.graph = 0 ;
         }
         if (PRE val.w.window) {
            S_display.Close_window(PRE val.w.window) ;
            PRE val.w.window = 0 ;
         }
         break ;
case EXPR_KEY:
         if (PRE val.key) {
            delete PRE val.key ;
            PRE val.key = 0 ;
         }
         break ;
default:
         break ;
      }
#ifdef DEBUG  // if we are debugging, set both pointers to -1
      PRE val.cons.car = (Expr *)(-1) ;
      PRE val.cons.cdr = (Expr *)(-1) ;
      PRE type = -1 ;
#else
      PRE val.cons.car = exprfreelist ;
      exprfreelist = THIS ;
#endif
   }
#ifndef CPLUSV2POINT0
   this = 0 ;
#endif
}

//  We need a default Null expression, since a `0' means no expression.
//  We also define |True| here, which is an expression equal to |#t|.

Expr null(Null, Null) ;
Expr true("#t") ;
Expr undefined("#u") ;
Expr badnum(7734) ;
Expr begin_lit("begin") ;
Expr lambda_lit("lambda") ;
Expr type_lit("type") ;
Expr quote_lit("quote") ;
Expr unquote_lit("unquote") ;
Expr unquote_splicing_lit("unquote_splicing") ;
Expr quasiquote_lit("quasiquote") ;

//  Displaying an expression is pretty routine.  We make some effort
//  to display it moderately well, but we're not overly concerned.
//  But first, we define the maximum length of an output line.

#define MAXLENGTH (75)

int Expr::display(FILE *f, int ilevel =0, int curpos =0) {
   char *temp = tempstring() ;
   const char *p ;
   int i ;
   int t ;
   Expr *q ;

   if (f==0)
      error("! NULL passed as file handle to display") ;
   switch (type) {
case EXPR_COMPLEX:
      if (curpos > ilevel) {
         fprintf(f, "\n") ;
         curpos = 0 ;
      }
      if (curpos >= 0)
         while (curpos < ilevel) {
            fprintf(f, " ") ;
            curpos++ ;
         }
      fprintf(f, "(") ;
      curpos++ ;
      for (q=this; q != Null;) {
         curpos = q->val.cons.car->display(f, ilevel+1, curpos) ;
         q = q->val.cons.cdr ;
         if (q != Null) {
            if (q->consp()) {
               fprintf(f, " ") ;
               curpos++ ;
            } else {
               fprintf(f, " . ") ;
               curpos += 3 ;
               curpos = q->display(f, ilevel, curpos) ;
               break ;
            }
         }
      }
      fprintf(f, ")") ;
      curpos++ ;
      break ;
case EXPR_ARRAY:
      if (curpos > ilevel) {
         fprintf(f, "\n") ;
         curpos = 0 ;
      }
      if (curpos >= 0)
         while (curpos < ilevel) {
            fprintf(f, " ") ;
            curpos++ ;
         }
      if (lambdamax) {
         fprintf(f, "{") ;
         curpos++ ;
         curpos = val.array.arr[0]->car()->display(f, ilevel+1, curpos) ;
         fprintf(f, " ") ;
         curpos++ ;
         curpos = val.array.arr[1]->display(f, ilevel+1, curpos) ;
         fprintf(f, "}") ;
         curpos++ ;
      } else {
         fprintf(f, "<") ;
         curpos++ ;
         for (i=0; i<val.array.len; i++) {
            if (i > 0) {
               fprintf(f, " ") ;
               curpos++ ;
            }
            curpos = val.array.arr[i]->display(f, ilevel+1, curpos) ;
         }
         fprintf(f, ">") ;
         curpos++ ;
      }
      break ;
case EXPR_LAMBDA:
      if (curpos > ilevel) {
         fprintf(f, "\n") ;
         curpos = 0 ;
      }
      while (curpos < ilevel) {
         fprintf(f, " ") ;
         curpos++ ;
      }
/*    fprintf(f, "[LAMBDA ") ;
      curpos += 8 ; */
      fprintf(f, "[LAMBDA (") ;
      curpos += 9 ;
      for (t=0, q=this->lambdaArgs(); q != Null; t++, q = q->cdr()) {
         if (t) {
            fprintf(f, " ") ;
            curpos++ ;
         }
         if (q->consp())
            curpos = q->car()->display(f, ilevel+9, curpos) ;
         else {
            fprintf(f, ". ") ;
            curpos = q->display(f, ilevel+9, curpos+2) ;
         }
      }
      fprintf(f, ") ") ;
      curpos += 2 ;
      curpos = this->val.lambda.SA[0]->display(f, ilevel+1, curpos) ;
      fprintf(f, "]") ;
      curpos++ ;
      break ;
case EXPR_FILE:
      p = "[FILE]" ;
      goto nextpart ;
case EXPR_SCOPE:
      p = "[HASH]" ;
      goto nextpart ;
case EXPR_GRAPH:
      p = "[GRAPH]" ;
      goto nextpart ;
case EXPR_NODE:
      p = "[NODE]" ;
      goto nextpart ;
case EXPR_WINDOW:
      p = "[WINDOW]" ;
      goto nextpart ;
case EXPR_KEY:
      p = "[KEY]" ;
      goto nextpart ;
case EXPR_BIND:
      p = "[BINDING]" ;
      goto nextpart ;
case EXPR_SCOPED:
      if (debugging('s')) {
         sprintf(temp, "%s{%d}", boundScopeEntry()->id(), scopeIndex()) ;
         goto temppart ;
      } else
         p = boundScopeEntry()->id() ;
      goto nextpart ;
case EXPR_LOCAL:
      if (debugging('s')) {
         sprintf(temp, "%s<%d>", boundScopeEntry()->id(), scopeIndex()) ;
         goto temppart ;
      } else
         p = boundScopeEntry()->id() ;
      goto nextpart ;
case EXPR_GLOBAL:
      p = globalScopeEntry()->id() ;
      goto nextpart ;
case EXPR_STRING:
      sprintf(temp, "%s", val.id) ;
      goto temppart ;
case EXPR_NUMBER:
      sprintf(temp, "%d", val.integer) ;
      goto temppart ;
case EXPR_WIRE:
      sprintf(temp, "W%d", val.wire.integer) ;
      goto temppart ;
case EXPR_PROC:
      sprintf(temp, "[PROC%d]", this->serial()) ;
      goto temppart ;
case EXPR_FORM:
      sprintf(temp, "[FORM%d]", this->serial()) ;
      goto temppart ;
case EXPR_ID:
      if (debugging('s')) {
         sprintf(temp, "|%s|", val.id) ;
         goto temppart ;
      } else
         p = val.id ;
      goto nextpart ;
case EXPR_EV_FORM:
      sprintf(temp, "[EVFORM%d]", this->serial()) ;
temppart:
      p = temp ;
nextpart:
      i = strlen(p) ;
      if (curpos + i > MAXLENGTH) {
         fprintf(f, "\n") ;
         for (t=0; t<ilevel; t++)
            fprintf(f, " ") ;
         curpos = ilevel ;
      }
      fprintf(f, "%s", p) ;
      curpos += i ;
      break ;
default:
      fprintf(stderr, ">>> <%lx:%d>\n", this, type) ;
      error("! some sort of illegal expression is trying to display.\n") ;
   }
   return(curpos) ;
}

//  Setting the `cdr' insures that we have a cdr to set.  One must be
//  very careful with these two expressions, as they totally bypass
//  the reference count mechanism.

void Expr::setcdr(Expr *b) {
   if (consp())
      val.cons.cdr = b ;
   else {
      if (this->display(stderr, 0, 0))
         fprintf(stderr, "\n") ;
      error("! tried to set the cdr of a node") ;
   }
}

//  Ditto with |setcar|.

void Expr::setcar(Expr *b) {
   if (type == EXPR_COMPLEX || type == EXPR_BIND)
      val.cons.car = b ;
   else {
      if (this->display(stderr, 0, 0))
         fprintf(stderr, "\n") ;
      error("! tried to set the car of a node") ;
   }
}

//  The |deepcopy| routine is pretty much the same as the delete routine.

Expr *Expr::deepcopy() {
   if (this == Null) {
      return(this) ;
   } else {
      if (type == EXPR_COMPLEX) {
         Expr *front = val.cons.car->deepcopy() ;
         Expr *back = val.cons.cdr->deepcopy() ;
         return new Expr(front, back) ;
      } else if (type == EXPR_LAMBDA || type == EXPR_BIND) {
         Expr *v = val.cons.car->deepcopy() ;
         return new Expr(v) ;
      } else if (type == EXPR_ID || type == EXPR_STRING) {
         return new Expr(val.id, type) ;
      } else
         return new Expr(val.integer) ;
   }
}

//  We routine the |id| only if the type is correct.

const char *Expr::id() {
   if (type == EXPR_ID)
      return(val.id) ;
   else {
      error2("Expression %s is not an id type", this->sername()) ;
      return(0) ;
   }
}

//  We routine the string only if the type is correct.

const char *Expr::string() {
   if (type == EXPR_STRING)
      return(val.id) ;
   else {
      error2("Expression %s is not a string type", this->sername()) ;
      return(0) ;
   }
}

//  We also have a routine to calculate the length of an expression list.
//  The length of |(a . b)| is considered to be 1.

int Expr::length() {
   Expr *r = this ;
   int len = 0 ;
   while (r != Null && r->consp()) {
      r = r->val.cons.cdr ;
      len++ ;
   }
   return(len) ;
}

//  Our function that pops a binding and deletes it is next.  Of course, if
//  someone else has `copied' this binding, then the thing won't actually be
//  deleted.

Expr *Expr::popBinding() {
   if (this == Null)
      error("! Popping a binding that doesn't exist") ;
   Expr *next = val.cons.cdr ;
   val.cons.cdr = Null ;
   this->remove() ;
   return next ;
}

ScopeEntry *Expr::scopeEntry(VOID) {
   if (boundp()) {
      return boundScopeEntry() ;
   } else if (globalp()) {
      return globalScopeEntry() ;
   } else
      return 0 ;
}

//  Finally, a `setbang' function.  Note that if what we are setbanging
//  to is a wire, then we actually create a wire if one hasn't already
//  been created, and then give it the value of the integer parameter.
//  (It must be an integer parameter or we report an error.)  This is
//  because |set!| is how we assign a new value to a wire, and |define|
//  is how we create that new wire.  Note how this is different than
//  binding (as in procedure calls), where the binding is analogous to
//  that for values.
//
//  For wires, we treat it `by reference'.  In reality, everything is
//  passed by reference, as an |Expr *|.  With wires, though, we munge
//  the value in the |Expr| structure itself, whereas with values we
//  allocate a new value if necessary.  This way, all current bindings
//  of wires see the new value.
//
//  For more wire magic, see the |Let| form in |Eval.cc|.

void Expr::setBang(Expr **scopeenv, Expr **localenv, Expr *s, int newp =0) {
   Expr *ret ;
   if (debugging('a')) {
      printf("Setting ") ; this->display(stdout) ; printf(" to ") ;
      s->display(stdout) ; printf("\n") ;
   }
   if (this == Null)
      return ;                          // use () as a default throw-away
   if (type == EXPR_GLOBAL) {
      ScopeEntry *se = val.global.se ;
      if (se->eType() == 0) {
         error("no type on global; kludging one on") ;
         se->setType(new Type(this)) ;
      }
      ret = se->valueNoCopy() ;
      if (se->eType()->wirep()) {
         int val = 0 ;
         if (s->nump())
            val = s->numv() ;
         else if (newp == 0 && s != Null)
            error("can only `set!' a wire to a value or Null. . .") ;
         s->remove() ;
         if (ret->wirep() == 0) {
            ret->remove() ;
            ret = new Expr(se->eType()->fixVal(val),
                                        se->eType()->typeWidth(), EXPR_WIRE) ;
            se->setValue(ret) ;
         } else {
            ret->setWireWidth(se->eType()->typeWidth()) ;
            ret->setWire(se->eType()->fixVal(val)) ;
         }
      } else {
         ret->remove() ;
         se->setValue(se->typeCast(s)) ;
      }
   } else if (boundp()) {
      ScopeEntry *se = val.bound.se ;
      Expr **v = ((type == EXPR_LOCAL) ? localenv : scopeenv)
                                                      + val.bound.integer ;
      if (newp) { // if we are from a |let| statement, we blow away any
         if (*v != Null) // old remnants from an earlier let statement
            (*v)->remove() ;
         *v = new Expr(Null, EXPR_BIND) ;
      }
      ret = (*v)->savalNoCopy() ;
      if (se->eType()->wirep()) {
         int val = 0 ;
         if (s->nump())
            val = s->numv() ;
         else if (newp == 0 && s != Null)
            error("can only `set!' a wire to a value . . .") ;
         s->remove() ;
         if (ret->wirep() == 0) {
            ret->remove() ;
            ret = new Expr(se->eType()->fixVal(val),
                                     se->eType()->typeWidth(), EXPR_WIRE) ;
            (*v)->setcar(ret) ;
         } else {
            ret->setWireWidth(se->eType()->typeWidth()) ;
            ret->setWire(se->eType()->fixVal(val)) ;
         }
      } else {
         ret->remove() ; // since ->saval() also copies . . .
         (*v)->setcar(se->typeCast(s)) ;
      }
   } else {
      this->display(stderr, 0, 0) ;
      error(": bad first arg for set or define") ;
   }
}

//  Our own storage management simply caches the freed |Expr|'s for later use.
//  We allocate |Expr|'s in big chunks to be easier on the system, both in
//  memory usage and in time.

static void Expr::makeExprBlock() {
   int i = 500 ;
   Expr *t = (Expr *)new char[sizeof(Expr) * i] ;
   exprfreelist = t ;
   while (--i > 0) {
      t->val.cons.car = t + 1 ;
      t++ ;
   }
   t->val.cons.car = 0 ;
}

//  Things are only slightly different between 2.0 and 1.2, for our
//  purposes.

Expr *exprfreelist ;
#ifdef CPLUSV2POINT0
void *Expr::operator new(long) {
   Expr *t ;
   if (exprfreelist == 0)
      exprfreelist->makeExprBlock() ;
   t = exprfreelist ;
   exprfreelist = exprfreelist->val.cons.car ;
   return (void *)t ;
}
#else
Expr *Expr::myexpralloc() {
   Expr *t ;
   if (this == 0) {
      if (exprfreelist == 0)
         exprfreelist->makeExprBlock() ;
      t = exprfreelist ;
      exprfreelist = exprfreelist->val.cons.car ;
      return t ;
   } else
      return this ;
}
#endif

//  Close the file if it's open.

void Expr::closefile() {
   if (filep() && val.file) {
      fclose(val.file) ;
      val.file = 0 ;
   }
}

//  Get the identifer, even if we've bound it.

const char *Expr::searchId() {
   if (type == EXPR_ID)
      return val.id ;
   else if (boundp())
      return val.bound.se->id() ;
   else if (type == EXPR_GLOBAL)
      return val.global.se->id() ;
   else
      return 0 ;
}

Expr *Expr::lambdaToClosure(Expr **scopeenv, Expr **localenv) {
   Expr **sa ;
   Expr *t = this->car() ;
   int level = t->car()->numv() ;
   t = t->cdr() ;
   int arity = t->car()->numv() ;
   t = t->cdr() ;
   int used = t->car()->numv() ;
   t = t->cdr() ;
   sa = new Expr *[level + 3] ;
   *sa++ = t->cdr()->copy() ;  // save arg list for deletion with lambda
   *sa++ = t->car()->copy() ;  // save hash for deletion with lambda
   t = new Expr(sa, (short)arity, (short)level, (short)used) ;
   *sa++ = this->cdr()->car()->copy() ;
   if (scopeenv) {
      scopeenv++ ;
      while (*scopeenv) {
         *sa++ = (*scopeenv++)->copy() ;
         level-- ;
      }
   }
   while (level > 1) {
      *sa++ = (*localenv++)->copy() ;
      level-- ;
   }
   *sa++ = 0 ;
   return t ;
}

Expr *Expr::namedLetToClosure(Expr **scopeenv, Expr **localenv) {
   Expr **sa ;
   Expr *t = this->car() ;
   int level = t->car()->numv() + 1 ;
   t = t->cdr() ;
   int arity = t->car()->numv() ;
   t = t->cdr() ;
   int used = t->car()->numv() ;
   t = t->cdr()->car() ;
   sa = new Expr *[level + 3] ;
   *sa++ = t->cdr()->copy() ;  // save arg list for deletion with lambda
   *sa++ = t->car()->copy() ;  // save hash for deletion with lambda
   t = new Expr(sa, (short)arity, (short)level, (short)used) ;
   *sa++ = this->cdr()->car()->copy() ;
   if (scopeenv) {
      scopeenv++ ;
      while (*scopeenv) {
         *sa++ = (*scopeenv++)->copy() ;
         level-- ;
      }
   }
   while (level > 2) {
      *sa++ = (*localenv++)->copy() ;
      level-- ;
   }
   *sa++ = new Expr(Null, EXPR_BIND) ;
   *sa++ = 0 ;
   return t ;
}

//  This function makes a new array.  We'd like to overload |Expr|, but
//  there are too many different types of |Expr|s whose construction
//  parameters are integers.

Expr *Expr::makeArray(int len) {
   Expr *p = new Expr(len, EXPR_ARRAY) ;

   p->val.array.len = len ;
   p->val.array.arr = new Expr *[len] ;
   while (len > 0)
      p->val.array.arr[--len] = Null ;
   p->lambdamax = 0 ;
   return p ;
}

//  This function sets the flag to treat the array as a structure and
//  only print specific information.

Expr *Expr::structArray(VOID) {
   lambdamax = 1 ;
}

//  These two functions access and set values in an array.

Expr *Expr::aref(int where) {
   if (type != EXPR_ARRAY) {
      error("expected array type in aref") ;
      return Null ;
   }
   if (where < 0 || where >= val.array.len) {
      error("array access out of bounds") ;
      return Null ;
   }
   return val.array.arr[where] ;
}

void Expr::aset(int where, Expr *what) {
   if (type != EXPR_ARRAY) {
      error("expected array type in aref") ;
      return ;
   }
   if (where < 0 || where >= val.array.len) {
      error("array access out of bounds") ;
      return ;
   }
   val.array.arr[where]->remove() ;
   val.array.arr[where] = what ;
}
