//  This is our main expression class.  We allow the following types:\:
//     complex (a list type)\:
//     number (an integer constant)\:
//     id (a pointer to a string)\:
//     lambda (a lambda expression)\:
//     binding (a binding type; just a mutable pointer to an Expr)\:
//     bound id (a scoped id that's been bound already)\:
//     global id (a global id that's already looked up)\:
//     procedure (a pointer to a procedure)\:
//     non-expand procedure (a pointer to a procedure that doesn't
//        want its arguments expanded)\:
//  Nothing else is allowed; that is a complete list.  Thus, we use a
//  union to represent it.

#include "Object.h"

//  Our general evaluation functions are all of this type.  (Using a
//  |typedef| saves a lot of typing.)

typedef class Expr *(evalftype)(Expr **args, int numargs) ;

//  We use the following enumeration to make things symbolic.

enum exprtype { EXPR_COMPLEX, EXPR_NUMBER, EXPR_WIRE, EXPR_FILE, EXPR_NODE,
                EXPR_GRAPH, EXPR_WINDOW, EXPR_SCOPE, EXPR_KEY, EXPR_ID,
                EXPR_ARRAY, EXPR_BIND, EXPR_LOCAL, EXPR_SCOPED, EXPR_GLOBAL,
                EXPR_STRING, EXPR_FORM, EXPR_EV_FORM, EXPR_LAMBDA,
                EXPR_PROC } ;

//  The flags we annotate our primitives with.

#define HARDPRIM (1)

//  For the |Expr| class, we do our own storage management.

extern Expr *exprfreelist ;

//  The actual class definition keeps getting larger.  Oh well.

class Expr : public Object {
   short type ; // we still have a `short' space here . . .
   short lambdamax ; // keep things long-aligned . . . use this sometime
   union {
      struct {
         Expr *car, *cdr ;
      } cons ;
      int integer ;
      struct {
         int integer ;
         int width ;
      } wire ;
      struct {
         class ScopeEntry *se ;
         int integer ;
      } bound ;
      char *id ;
      struct {
         evalftype *proc ;
         int flags ;
      } builtin ;
      struct {
         class ScopeEntry *se ;
         const char *id ;
      } global ;
      struct {
         short low, arity ;
         Expr **SA ;
      } lambda ;
      struct {
         int len ;
         Expr **arr ;
      } array ;
      FILE *file ;
      class Scope *scope ;
      class Key *key ;
      class HWSG *graph ;
      class Node *node ;
      struct {
         Expr *graph ;
         int window ;
      } w ;
   } val ;
   int refcount ;
public:

//  Some standard Lisp-ish functions.

   Expr *car(VOID) ;
   Expr *fastcar(VOID) { return val.cons.car ; } // ref Tracy Chapman
   Expr *cdr(VOID) ;
   Expr *fastcdr(VOID) { return val.cons.cdr ; }
   int length(VOID) ;
   void setcdr(Expr *t) ;
   void setcar(Expr *t) ;

//  What are we?

   int consp(VOID) { return type == EXPR_COMPLEX ; }
   int nump(VOID) { return type == EXPR_NUMBER ; }
   int idp(VOID) { return type == EXPR_ID ; }
   int stringp(VOID) { return type == EXPR_STRING ; }
   int procp(VOID) { return type >= EXPR_LAMBDA ; }
   int evprocp(VOID) { return type >= EXPR_EV_FORM ; }
   int formp(VOID) { return type == EXPR_FORM ; }
   int evformp(VOID) { return type == EXPR_EV_FORM ; }
   int lambdap(VOID) { return type == EXPR_LAMBDA ; }
   int boundp(VOID) { return type == EXPR_SCOPED || type == EXPR_LOCAL ; }
   int scopedp(VOID) { return type == EXPR_SCOPED ; }
   int localp(VOID) { return type == EXPR_LOCAL ; }
   int filep(VOID) { return type == EXPR_FILE ; }
   int scopep(VOID) { return type == EXPR_SCOPE ; }
   int globalp(VOID) { return type == EXPR_GLOBAL ; }
   int graphp(VOID) { return type == EXPR_GRAPH ; }
   int nodep(VOID) { return type == EXPR_NODE ; }
   int windowp(VOID) { return type == EXPR_WINDOW ; }
   int wirep(VOID) { return type == EXPR_WIRE ; }
   int keyp(VOID) { return type == EXPR_KEY ; }
   int arrayp(VOID) { return type == EXPR_ARRAY ; }
   int etype(VOID) { return type ; }

//  How can we get the values?

   const char *id(VOID) ;
   const char *searchId(VOID) ;
   const char *string(VOID) ;
   evalftype *evaluatef(VOID) { return val.builtin.proc ; }
   int hardprimp(VOID) { return val.builtin.flags & HARDPRIM ; }
   class ScopeEntry *globalScopeEntry(VOID) { return val.global.se ; }
   class ScopeEntry *boundScopeEntry(VOID) { return val.bound.se ; }
   class ScopeEntry *scopeEntry(VOID) ;
   int scopeIndex(VOID) { return val.bound.integer ; }
   Expr *saval(VOID) { return val.cons.car->copy() ; }
   Expr *savalNoCopy(VOID) { return val.cons.car ; }
   int numv(VOID) { return val.integer ; }
   int formv(VOID) { return val.integer ; }
   int wireValue(VOID) { return val.wire.integer ; }
   int wireWidth(VOID) { return val.wire.width ; }
   int lambdaLow(VOID) { return val.lambda.low ; }
   int lambdaArity(VOID) { return val.lambda.arity ; }
   int lambdaMax(VOID) { return lambdamax ; }
   int refCount() { return refcount ; }
   int arrayLength() { return val.array.len ; }
   Expr **arrayPointer() { return val.array.arr ; }
   class Key *keyv() { return val.key ; }
   Expr *callPrim(Expr **args, int numargs)
                                   { return evaluatef()(args, numargs) ; }
   Expr *lambdaExpr(VOID) { return val.lambda.SA[0] ; }
   Expr **lambdaSA(VOID) { return val.lambda.SA ; }
   Expr *lambdaArgs(VOID) { return val.lambda.SA[-2] ; }
   Expr *lambdaHash(VOID) { return val.lambda.SA[-1] ; }
   Expr *lambdaToClosure(Expr **scopeenv, Expr **localenv) ;
   Expr *namedLetToClosure(Expr **scopeenv, Expr **localenv) ;
   Expr *aref(int where) ;
   void aset(int where, Expr *what) ;
   Expr *quotifyList(VOID) ;
   class HWSG *exprGraph(VOID) { return val.graph ; }
   class Node *exprNode(VOID) { return val.node ; }
   int exprWindow(VOID) { return val.w.window ; }
   void *valAddress(VOID) { return (void *)&val ; }
   void setWire(int i) { val.wire.integer = i ; }
   void setWireWidth(int w) { val.wire.width = w ; }
   FILE *file(VOID) { return val.file ; }
   void closefile(VOID) ;
   class Scope *scope(VOID) { return val.scope ; }

//  Constructors and destructors.  If we are using a C++ 2.0 or later
//  compiler, we can easily override the operator |new| function;
//  otherwise we have to do other magic.
//
//  We also define a |remove| member.  We do this to get around the
//  destructor messiness when we don't actually delete the object;
//  we don't want the base class destructor to be called.  This also
//  allows us to inline the refcount decrement and test for some
//  additional speed.

#ifdef CPLUSV2POINT0
#define EXPRALLOC refcount = 1
   void *operator new(long) ;
   void operator delete(void *) ;
   ~Expr(VOID) {}
#else
   Expr *myexpralloc(VOID) ;
   ~Expr(VOID) ;
#define EXPRALLOC this = myexpralloc() ; refcount = 1
#endif
   void makeExprBlock(void) ;
   void remove() { if (--refcount <= 0) delete this ; }
   Expr(const char *s, int t =EXPR_ID)
      { EXPRALLOC ; type = t ; val.id = newstring(s) ; }
   Expr(Expr *ar, int t =EXPR_LAMBDA)
      { EXPRALLOC ; type = t ; val.cons.car = ar ; }
   Expr(FILE *f) { EXPRALLOC ; type = EXPR_FILE ; val.file = f ; }
   Expr(class Scope *s) { EXPRALLOC ; type = EXPR_SCOPE ; val.scope = s ; }
   Expr(int num, int t =EXPR_NUMBER)
      { EXPRALLOC ; type = t ; val.integer = num ; }
   Expr(int num, int w, int t)
     { EXPRALLOC ; type = t ; val.wire.integer = num ; val.wire.width = w ; }
   Expr(Expr *ar, Expr *dr, int t =EXPR_COMPLEX)
      { EXPRALLOC ; type = t ; val.cons.car = ar ; val.cons.cdr = dr ; }
   Expr(class ScopeEntry *se)
      { EXPRALLOC ; type = EXPR_GLOBAL ; val.global.se = se ; }
   Expr(class ScopeEntry *se, int lev, int t =EXPR_SCOPED)
     { EXPRALLOC ; type = t ; val.bound.se = se ; val.bound.integer = lev ; }
   Expr(Expr **e, short a, short l, short u)
      { EXPRALLOC ; type = EXPR_LAMBDA ; val.lambda.SA = e ; lambdamax = u ;
            val.lambda.arity = a ; val.lambda.low = l ; }
   Expr(evalftype *f, int flags =0) { EXPRALLOC ; type = EXPR_PROC ;
                    val.builtin.proc = f ; val.builtin.flags = flags ; }
   Expr(class HWSG *g) { EXPRALLOC ; type = EXPR_GRAPH ; val.graph = g ; }
   Expr(class Node *n) { EXPRALLOC ; type = EXPR_NODE ; val.node = n ; }
   Expr(int w, Expr *g)
    { EXPRALLOC ; type = EXPR_WINDOW ; val.w.graph = g ; val.w.window = w ; }
   Expr(class Key *d) { EXPRALLOC ; type = EXPR_KEY ; val.key = d ; }

//  Some more complicated functions.

   int display(FILE *f, int ilevel =0, int curpos =0) ;
   const char *carid(VOID) ;
   Expr *macroexpand(int level =0) ;
   Expr *alphaBind(class Scope *scope, int *maxuse, int nv =1,
                                               int clev =1, int qqlev =0) ;
   Expr *eval(int numlocals) ;
   Expr *copy(VOID) { refcount++ ; return this ; }
   Expr *deepcopy(VOID) ;
   Expr *forever(VOID) { refcount = 2000000000 ; return this ; }
   Expr *popBinding(VOID) ;
   Expr *makeArray(int len) ;
   Expr *structArray(VOID) ;
   void setBang(Expr **scopeenv, Expr **localenv, Expr *exp, int newp =0) ;
   virtual const char *isa(VOID) { return "Expr" ; }
} ;

//  Finally, we make our |Null| and |True| expressions available.

extern Expr null, true, undefined, badnum ;
extern Expr begin_lit, lambda_lit, type_lit, quote_lit,
            unquote_lit, unquote_splicing_lit, quasiquote_lit ;
#define Null (&null)
#define True (&true)
#define Undefined (&undefined)
#define BadNum (&badnum)
#define Literal_begin (&begin_lit)->copy()
#define Literal_lambda (&lambda_lit)->copy()
#define Literal_type (&type_lit)->copy()
#define Literal_quote (&quote_lit)->copy()
#define Literal_unquote (&unquote_lit)->copy()
#define Literal_unquote_splicing (&unquote_splicing_lit)->copy()
#define Literal_quasiquote (&quasiquote_lit)->copy()
