//  The implementation of the graph class is straightforward.
//  Most everything was taken care of in the declaration of the class.

#include "Graph.h"
#include "Expr.h"
#include "layout.h"
#include "display.h"

//  Our |Node| constructor must allocate arrays to hold the pointers to
//  the lists, as well as the lists themselves.

static const char *defaultname = "no name" ;

Node::Node(int incount =1, int outcount =1) {
   inarity = incount ;
   outarity = outcount ;
   if (incount + outcount) {
      in = new List *[incount + outcount] ;
      out = in + incount ;
   }
   int i ;
   for (i=0; i<incount; i++)
      in[i] = new List ;
   for (i=0; i<outcount; i++)
      out[i] = new List ;
   savedName = defaultname ;
}

Node::Node(const char *s) {
   inarity = 1 ;
   outarity = 1 ;
   in = new List *[2] ;
   out = in + 1 ;
   in[0] = new List ;
   out[0] = new List ;
   savedName = newstring(s) ;
}

//  Deleting a |Node| is the inverse of the above code.

Node::~Node() {
   int i ;
   for (i=0; i<inarity; i++)
      delete in[i] ;
   for (i=0; i<outarity; i++)
      delete out[i] ;
   if (inarity + outarity)
      delete[inarity + outarity] in ;
   if (savedName != defaultname)
      delete savedName ;
}

//  These two functions return the total number of input or output
//  edges.

int Node::inCount(VOID) {
   int i, s ;
   for (i=0, s=0; i<inArity(); i++)
      s += inEdges(i)->length() ;
   return s ;
}

int Node::outCount(VOID) {
   int i, s ;
   for (i=0, s=0; i<outArity(); i++)
      s += outEdges(i)->length() ;
   return s ;
}

//  The constructor for the edge class adds the edge to both node lists.

Edge::Edge(Node *t, Node *h, int inplace =0, int outplace =0) {
   inpos = inplace ;
   outpos = outplace ;
   headPtr = h ; tailPtr = t ;
   t->out[inpos]->append(this) ;
   h->in[outpos]->append(this) ;
}

//  The destructor cleans up the lists.  Note that we need to find and remove,
//  rather than just remove, since we don't necessarily know where on the
//  list the item occurs.

Edge::~Edge() {
   tailPtr->out[inpos]->findAndRemove(this) ;
   headPtr->in[outpos]->findAndRemove(this) ;
}

//  Our graph destructor has to be careful to delete all of the edges
//  and nodes listed.  The macro |delete_a_list_of| is defined in
//  |List.h|.

Graph::~Graph() {
   if (layout)
      layout_clean(layout) ;
   deleteAListOf(Edge, edges) ;
   deleteAListOf(Node, nodes) ;
   delete nodes ;
   delete edges ;
}

//  We also want two functions that will remove an edge or node from a
//  graph cleanly, including deleting any references to that edge or
//  node.  First, we deal with deleting an edge, since that is
//  moderately straightforward.  We have to delete the edge from the
//  list of edges we are maintaining.
   
void Graph::deleteEdge(Edge *e) {
   edges->findAndRemove(e) ;
   delete e ;
}

//  When deleting a node, we also delete all edges to that node.  We return
//  the number of such edges deleted, since often we will delete the edges
//  first and if any remain we have a problem.

int Graph::deleteNode(Node *e) {
   int removed = 0 ;
   int i ;
   for (i=e->inArity()-1; i>=0; i--)
      while (e->in[i]->length()) {
         this->deleteEdge((Edge *)(e->in[i]->firstEl())) ;
         removed++ ;
      }
   for (i=e->outArity()-1; i>=0; i--)
      while (e->out[i]->length()) {
         this->deleteEdge((Edge *)(e->out[i]->firstEl())) ;
         removed++ ;
      }
   nodes->findAndRemove(e) ;
   return(removed) ;
}

//  We define the various accessor functions here to build a window, and finally
//  add a method to the |Expr| class called |makeWindow|.

static Graph *currgraph ;

void fillnodearray(void **nodearr) {
   GraphNodeIterator gei(*currgraph) ;
   Node *n = 0 ;
   int i = 0 ;

   while (n=gei()) {
      n->setIndex(i++) ;
      *nodearr++ = n ;
   }
   if (i != currgraph->nodeList()->length())
      error("! consistency error when filling node array") ;
}

void filledgearray(void *node, int *incnt, int *outcnt,
                           int *inarity, int *outarity, int *succarr) {
   Node *n = (Node *)node ;
   int i ;
   int addin = 0, addout = 0 ;
   Edge *e ;

   for (i=0; i<n->outArity(); i++) {
      ListIterator li(n->outEdges(i)) ;
      while (e = (Edge *)(li())) {
         *succarr++ = e->head()->getIndex() ;
         addout++ ;
      }
   }
   for (i=0; i<n->inArity(); i++)
      addin += n->inEdges(i)->length() ;
   *outcnt = addout ;
   *incnt = addin ;
   *inarity = n->inArity() ;
   *outarity = n->outArity() ;
}

void filledgeptrs(void *node, void **varr) {
   Node *n = (Node *)node ;
   int i ;
   Edge *e ;

   for (i=0; i<n->outArity(); i++) {
      ListIterator li(n->outEdges(i)) ;
         while (e = (Edge *)(li()))
            *varr++ = e ;
   }
}

void filledgeio(void *node, short *io) {
   Node *n = (Node *)node ;
   int i ;
   Edge *e ;

   for (i=0; i<n->outArity(); i++) {
      ListIterator li(n->outEdges(i)) ;
         while (e = (Edge *)(li())) {
            *io++ = e->inPos() ;
            *io++ = e->outPos() ;
         }
   }
}

int getwidth(void *n) {
   return ((Node *)n)->width() ;
}

int getheight(void *n) {
   return ((Node *)n)->height() ;
}

void getoffsets(void *node, short *in, short *out) {
   Node *n = (Node *)node ;
   int i, k, w ;

   w = getwidth(n) + 8 ;
   k = n->inArity() ;
   if (k > 1)
      for (i=0; i<k; i++)
         *in++ = ((2 * i + 1 - k) * w) / (k + k) ;
   k = n->outArity() ;
   if (k > 1)
      for (i=0; i<k; i++)
         *out++ = ((2 * i + 1 - k) * w) / (k + k) ;
}

void Node::draw(int x, int y) {
   int w = getwidth(this) ;
   int h = getheight(this) ;
   set_color(1) ;
   draw_box(x-w/2, y-h/2, w, h) ;
   label(x, y, name()) ;
}

void drawnode(void *n, int x, int y) {
   ((Node *)n)->draw(x, y) ;
}

void Edge::draw(int x1, int y1, int x2, int y2) {
   set_color(1) ;
   draw_line(x1, y1, x2, y2) ;
}

void drawedge(void *e, int x1, int y1, int x2, int y2, char) {
   ((Edge *)e)->draw(x1, y1, x2, y2) ;
}

struct layout_adjust_params lap = { 960, 0, 0, 12, 5, 17, 7, 1 } ;

//  The following routines are for named nodes, so we can easily
//  and dynamically create arbitrary graphs with HWSG.

Node *Graph::findNamedNode(const char *s) {
   GraphNodeIterator gei(*this) ;
   Node *n = 0 ;
   while (n=gei())
      if (strcmp(s, n->name())==0)
         return n ;
   return 0 ;
}

void Graph::addNamedNode(const char *s) {
   Node *n = findNamedNode(s) ;
   if (n==0)
      n = addNode(new Node(s)) ;
}

void Graph::deleteNamedNode(const char *s) {
   Node *n = findNamedNode(s) ;
   if (n)
      deleteNode(n) ;
}

Edge *Graph::findNamedEdge(const char *t, const char *h) {
   Node *tail = findNamedNode(t) ;
   Node *head = findNamedNode(h) ;
   Edge *e ;
   int i ;
   if (tail && head) {
      for (i=0; i<tail->outArity(); i++) {
         ListIterator li(tail->outEdges(i)) ;
            while (e = (Edge *)(li()))
               if (e->head() == head)
                  return e ;
      }
   }
   return 0 ;
}

void Graph::addNamedEdge(const char *t, const char *h) {
   Node *tail = findNamedNode(t) ;
   Node *head = findNamedNode(h) ;
   Edge *e ;
   int i ;
   if (tail && head) {
      for (i=0; i<tail->outArity(); i++) {
         ListIterator li(tail->outEdges(i)) ;
            while (e = (Edge *)(li()))
               if (e->head() == head)
                  return ;
      }
   }
   addEdge(new Edge(tail, head)) ;
}

void Graph::deleteNamedEdge(const char *t, const char *h) {
   Edge *e = findNamedEdge(t, h) ;
   if (e)
      deleteEdge(e) ;
}

Expr *Graph::makeWindow(Expr *g =0) {
   int width, height ;
   int w ;

   if (debugging('g'))
      lo_debug = 1 ;
   else
      lo_debug = 0 ;
   if (g)
      currgraph = (Graph *)(g->exprGraph()) ;
   else {
      currgraph = this ;
      g = Null ;
   }
   layout = layout_graph(nodes->length(), fillnodearray, filledgearray,
      filledgeptrs, filledgeio, 0) ;
   layout_adjust(layout, getwidth, getheight, getoffsets, &lap,
      &width, &height) ;
   w = open_window(width, height) ;
   layout_draw(layout, drawnode, drawedge) ;
   flush_window() ;
   return new Expr(w, g->copy()) ;
}
