//  This file defines a generalized multigraph class.  We want to be able
//  to construct and modify a general graph, and associate any type of
//  information with any of the nodes or edges.  We use the list class.
//
//  These graphs need not be connected.  They are directed, but should be
//  easy to treat as undirected.

#include "List.h"
#include "display.h"

//  Our first class is the node class.  We maintain lists of pointers to
//  all outgoing and incoming edges.  Note that these lists should only
//  contain |Edge| types.

#define MAX_NODE_HEIGHT (string_height()+3)
#define DEFAULT_NODE_WIDTH (string_height()*3)
#define LENTOWID (string_width(name())+6)

class Node : public Object {
   friend class Graph ;
   friend class Edge ;
   int inarity, outarity ;
   int index ;                 // for layout purposes
   int misc ;                  // for whoever wants to use it
   const char *savedName ;
   List **in, **out ;          // our incoming and outgoing edges.
public:
   virtual const char *isa() { return "Node" ; }
   Node(int incount =1, int outcount =1) ;
   Node(const char *s) ;
   ~Node(VOID) ; // we would really like to inline this simple one.
   virtual const char *name(VOID) { return savedName ; }
   int inArity(VOID) { return inarity ; }
   int outArity(VOID) { return outarity ; }
   int inCount(VOID) ;
   int outCount(VOID) ;
   int getIndex(VOID) { return index ; }
   void setIndex(int n) { index = n ; }
   List *inEdges(int i) { return in[i] ; }
   List *outEdges(int i) { return out[i] ; }
   virtual void draw(int x, int y) ;
   virtual int height(VOID) { return MAX_NODE_HEIGHT ; }
   virtual int width(VOID) { return LENTOWID ; }
} ;

//  When we create an edge, we must have two nodes to attach it to.  The
//  constructor (in the attached file) adds it to the lists as appropriate.
//  Our |Edge| also contains information as to what position in the |Node|
//  it is added to, so we can easily locate it in the |Node|'s input and
//  output edge lists.

class Edge : public Object {
   friend class Graph ;
   Node *headPtr, *tailPtr ;    // the nodes at the head and tail of the edge
   int misc ;                   // for whoever wants to use it
   int inpos, outpos ;
public:
   Edge(Node *t, Node *h, int inplace =0, int outplace =0) ;
   ~Edge(VOID) ;
   virtual const char *isa(VOID) { return "Edge" ; }
   Node* head(VOID) { return headPtr ; }
   Node* tail(VOID) { return tailPtr ; }
   int inPos(VOID) { return inpos ; }
   int outPos(VOID) { return outpos ; }
   virtual void draw(int x1, int y1, int x2, int y2) ;
} ;

//  And finally, the whole burrito.  We maintain a list of edges and
//  nodes for easy iteration and deletion.
//
//  The general way to use this class is to create a graph with the
//  class constructor.  Then, each node is added, followed or
//  intermixed with the edges.  Both nodes must be added before an edge
//  is added.  When a node is created, the pointers you pass must be
//  unique because they are what is used to match up edges.

class Graph : public Object {
   List *nodes, *edges ;         // all of our nodes and edges are here
   void *layout ;
public:
   Node *addNode(Node *a) { return((Node*)nodes->append(a)) ; }
   Edge *addEdge(Edge *e) { return((Edge*)edges->append(e)) ; }
   Node *createNode(int incount =1, int outcount =1)
      { return ((Node *)nodes->append(new Node(incount, outcount))) ; }
   Edge *createEdge(Node *t, Node *h, int inplace =0, int outplace =0)
      { return ((Edge *)edges->append(new Edge(t, h, inplace, outplace))) ; }
   int deleteNode(Node *n) ;
   void deleteEdge(Edge *e) ;

//  The following handle named nodes.

   Node *findNamedNode(const char *s) ;
   void addNamedNode(const char *s) ;
   void deleteNamedNode(const char *s) ;
   Edge *findNamedEdge(const char *t, const char *h) ;
   void addNamedEdge(const char *t, const char *h) ;
   void deleteNamedEdge(const char *t, const char *h) ;

//  Now back to HWSG-style things.

   virtual const char *isa(VOID) { return "Graph" ; }
   List *nodeList(VOID) { return nodes ; }
   List *edgeList(VOID) { return edges ; }
   Expr *makeWindow(Expr *g =0) ;
   void *laidOut(VOID) { return layout ; }
   int searchForwardLessMisc(Node *from, Node *to, Edge *toignore = 0) ;
   int removeAllPathsFrom(Node *from, Node *to, class IntSet *is) ;
   struct reachStruct *interpret(class IntSet *marking, struct reachStruct **,
      Graph *underConstruction, int *tokencount) ;
   Graph *reachabilityGraph(VOID) ;
   Edge *redundantEdge(VOID) ;
   Node *IOConnected(VOID) ;
   Graph(VOID) { nodes = new List ; edges = new List ; layout = 0 ; }
   ~Graph(VOID) ;
} ;

//  We need a couple of iterators which are trivially derived
//  from the list iterator using the macros provided in |List.h|.

defineListIterator(GraphEdgeIterator, Edge*, Graph& g, g.edgeList())
defineListIterator(GraphNodeIterator, Node*, Graph& g, g.nodeList())
