//  Our implementation of HWSG constructs the graphs from lambda expressions.

#include "HWSG.h"

//  Our creation simply initializes a couple of things.

HWSG::HWSG() {
   who = 0 ;
   outputNode = 0 ;
}

//  Our deletion simply cleans a few things up.

HWSG::~HWSG() {
   if (who)
      who->remove() ;
}

HWSG::HWSG(class Expr *e) {
   outputNode = 0 ;
   who = e->copy() ; // also insures that we don't delete lambda accidentally
   if (! e->lambdap())
      error(" can't build a HWSG from a non-lambda.") ;
   else {

//  We allocate an array to hold pointers to our input values; we also
//  initialize these elements.  Finally, we just `addExpr' the body,
//  passing all relevant information.

      int inarity = e->lambdaArity() ;
      int max = e->lambdaMax() ;
      int i ;
      Node **Inputs = new Node *[max] ;
      Expr *argList = e->lambdaArgs() ;
      Place *pl = addPlace(new Place()) ;
      for (i=0; i<inarity; i++, argList = argList->cdr())
         Inputs[i] = (Node *)addInputNode(new InputNode(
                                         argList->car()->boundScopeEntry())) ;
      for (; i<max; i++)
         Inputs[i] = 0 ;
      outputNode = addExpr(e->lambdaExpr(), Inputs, 0, &pl) ;
      delete Inputs ;
   }
}

//  This function adds a single expression to our graph.  It is recursively
//  called, starting with the lambda expression body and working `upwards'.
//  This is the function that does the main work, keying off the function
//  parameter.

Node *HWSG::addExpr(Expr *e, Node **Is, Node **Close, Place **pl) {
   if (debugging('h')) {
      e->display(stdout, 0, 0) ;
      printf("\n") ;
   }
   Node *r = 0 ;
   if (e->consp()) {  // if it's a cons, we have an expression to evaluate.
      Expr *t = e->car() ;

//  Our first case is dealing with locally bound procedures.  A procedure can
//  be locally bound by being passed in (illegal at the moment), by being
//  locally bound by `let' (also illegal at the moment) or by being the
//  identifier in a named let (which is the common case we need to properly
//  handle.)
//
//  For functions of no argument, it's very simple; it's just a `goto'.
//  Dealing with the arguments in a nice data-flow way isn't so easy.
//  We'll probably kludge some weird `merge' dataflow node to handle this.

      if (t->boundp()) { // it's a local call.  it's most likely a loop.
         Node *testnode ;
         if (t->localp()) {
            testnode = Is[t->scopeIndex()] ;
         } else {
            if (Close == 0) {
               error("no legitimate closure") ;
               testnode = 0 ;
            } else
               testnode = Close[t->scopeIndex()] ;
         }
         if (testnode && strcmp(testnode->isa(), "Transition")==0) {
            addPetriEdge((Node *)*pl, testnode, 0, 0) ;
            *pl = 0 ;
            r = 0 ;
         } else {
            error("I encountered a bound function I didn't like") ;
            e->display(stderr) ;
            fprintf(stderr, "\n%d %d %d\n", Is, Close, t->scopeIndex()) ;
            r = 0 ;
         }
      } else if (t->globalp()) { // it's a global.  Thank God.  look it up.
         ScopeEntry *se = t->globalScopeEntry() ;
         Expr *v = se->valueNoCopy() ;
         if (v->procp()) {

//  If the car of the complex expression is a procedure call that expands
//  its arguments directly, we just generate a call node.  First, we build
//  the actual call node.  Then, we build the arguments, connecting them as
//  we encounter them.  But we check for a few cases that should be handled
//  specially.
//
//  |Void| is very simple; we simply return 0.

            const char *id = se->id() ;
            if (strcmp(id, "void")==0)
               r = 0 ;

//  For the value keyword, we have to `observe' the value of a wire.
//  We introduce a transition for this event.

            else if (strcmp(id, "value")==0) {
               Transition *nt = addTransition(new Transition()) ;
               Place *np = addPlace(new Place()) ;
               addPetriEdge((Node *)*pl, (Node *)nt, 0, 0) ;
               addPetriEdge((Node *)nt, (Node *)np, 0, 0) ;
               *pl = np ;
               r = addExpr(e->cdr()->car(), Is, Close, pl) ;
            } else {
               Node *callnode ;
               if (v->lambdap())
                  callnode = (Node *)addSubroutineNode(
                              new SubroutineNode(se, e->cdr()->length())) ;
               else
                  callnode = (Node *)addPrimitiveNode(
                              new PrimitiveNode(se, e->cdr()->length())) ;
               Expr *arg ;
               int i ;
               for (i=0, arg = e->cdr(); arg != Null; arg = arg->cdr(), i++)
                   addDFEdge(addExpr(arg->car(), Is, Close, pl),
                                                             callnode, 0, i) ;
               r = callnode ;
            }
         } else if (v->formp()) { // otherwise, handle other cases.

//  This is where we check for and handle things like `if', `let',
//  `set!', and other control constructs.

            const char *id = se->id() ;
            Node *callnode ;

//  The |if| is treated just like a call node as far as dataflow is
//  concerned.  But we add an |IfTransition| for control flow.

            if (strcmp(id, "if")==0) {
               callnode = (Node *)addIfNode(new IfNode()) ;
               IfTransition *ifTrans = addIfTransition(new IfTransition()) ;
               Place *tBranch = addPlace(new Place()) ;
               Place *fBranch = addPlace(new Place()) ;
               Place *jPlace = addPlace(new Place()) ;
               Transition *tTransit = addTransition(new Transition()) ;
               Transition *fTransit = addTransition(new Transition()) ;
               Expr *arg = e->cdr() ;
               Node *iftest = addExpr(arg->car(), Is, Close, pl) ;
               addDFEdge(iftest, (Node *)ifTrans, 0, 0) ;
               addDFEdge(iftest, callnode, 0, 0) ;
               addPetriEdge((Node *)*pl, (Node *)ifTrans, 0, 1) ;
               addPetriEdge((Node *)ifTrans, (Node *)tBranch, 0, 0) ;
               addPetriEdge((Node *)ifTrans, (Node *)fBranch, 1, 0) ;
               arg = arg->cdr() ;
               addDFEdge(addExpr(arg->car(), Is, Close, &tBranch),
                                                            callnode, 0, 1) ;
               arg = arg->cdr() ;
               addDFEdge(addExpr(arg->car(), Is, Close, &fBranch),
                                                            callnode, 0, 2) ;
               addPetriEdge((Node *)tBranch, (Node *)tTransit) ;
               addPetriEdge((Node *)fBranch, (Node *)fTransit) ;
               addPetriEdge((Node *)tTransit, (Node *)jPlace) ;
               addPetriEdge((Node *)fTransit, (Node *)jPlace) ;
               *pl = jPlace ;
               r = callnode ;

//  For |set!|, we check to see if we are setting a wire or a value.  If
//  it's a wire, we introduce the necessary transition to mark a change
//  on the wire.  If it's a bound value, we munge the |Inputs| array with
//  the new value.  If it's a global (which will happen), we complain;
//  this is a bug that we need to fix eventually.

            } else if (strcmp(id, "set!")==0) {
               Node *ret = addExpr(e->cdr()->cdr()->car(), Is, Close, pl) ;
               Expr *lvalue = e->cdr()->car() ;
               ScopeEntry *se = 0 ;
               if (lvalue->boundp())
                  se = lvalue->boundScopeEntry() ;
               else if (lvalue->globalp())
                  se = lvalue->globalScopeEntry() ;
               else
                  error("you can't set! a non-variable!") ;
               if (se) {
                  Type *ty = se->eType() ;
                  if (ty->wirep()) {
                     Transition *trans = addTransition(new Transition()) ;
                     Place *place = addPlace(new Place()) ;
                     addPetriEdge((Node *)*pl, (Node *)trans, 0, 0) ;
                     addPetriEdge((Node *)trans, (Node *)place, 0, 0) ;
                     *pl = place ;
                  } else if (lvalue->boundp())
                     Is[lvalue->scopeIndex()] = ret ;
                  else
                     error("I don't handle setting global values yet, sorry") ;
               }
               r = 0 ;

//  For |begin|, we evaluate each expression in turn and only return the
//  last.  If there are none, or the last returns |void|, we return |0|.
//  But there is a sequence point from statement to statement, so we
//  construct new transitions as we go if necessary.

            } else if (strcmp(id, "begin")==0) {
               Expr *l ;
               for (l=e->cdr(); l != Null; l = l->cdr())
                  r = addExpr(l->car(), Is, Close, pl) ;

//  Handling `let' is not difficult; we simply extend the |Is| array
//  with the expressions, and then expand the body.

            } else if (strcmp(id, "let")==0) {
               Expr *l=e->cdr()->car()->car()->cdr() ;
               Expr *t=e->cdr()->car()->cdr() ;
               for (; l != Null; l = l->cdr(), t = t->cdr())
                  Is[l->car()->scopeIndex()] = addExpr(t->car(),
                                                             Is, Close, pl) ;
               r = addExpr(e->cdr()->cdr()->car(), Is, Close, pl) ;
            } else if (strcmp(id, "namedlet")==0) {
               Expr *l = e->cdr()->car() ;
               int low = l->car()->numv() ;
               Node **newClose = new Node *[low + 2] ;
               int i = 1, j ;
               if (Close)
                  for (; Close[i]; i++)
                     newClose[i] = Close[i] ;
               for (j = 0; i < low; i++, j++)
                  newClose[i] = Is[j] ;
               l = l->cdr() ;
               int inarity = l->car()->numv() ;
               l = l->cdr() ;
               int max = l->car()->numv() ;
               l = l->cdr()->cdr() ;
               Node **Inputs = new Node *[max - low] ;
               Transition *nt = addTransition(new Transition()) ;
               Place *np = addPlace(new Place()) ;
               addPetriEdge((Node *)*pl, (Node *)nt, 0, 0) ;
               addPetriEdge((Node *)nt, (Node *)np, 0, 0) ;
               *pl = np ;
               newClose[i++] = (Node *)nt ;
               newClose[i] = 0 ;
               for (i=0; i<inarity; i++, l = l->cdr())
                  Inputs[i] = addExpr(l->car(), Is, Close, pl) ;
               r = addExpr(e->cdr()->cdr()->car(), Inputs, newClose, pl) ;
               delete Inputs ;
               delete Close ;
            } else {
               callnode = (Node *)addPrimitiveNode(
                           new PrimitiveNode(se, e->cdr()->length())) ;
               Expr *arg ;
               int i ;
               for (i=0, arg = e->cdr(); arg != Null; arg = arg->cdr(), i++)
                   addDFEdge(addExpr(arg->car(), Is, Close, pl),
                                                             callnode, 0, i) ;
               r = callnode ;
            }
         } else { // not a proc, complain.
            error("Expression in function position not currently a procedure") ;
            r = (Node *)addConstantNode(new ConstantNode(7734)) ;
         }
      } else {
         experror("Illegal expression in function position", e) ;
         r = (Node *)addConstantNode(new ConstantNode(7734)) ;
      }
   } else if (e->nump()) {   // case of integer constant
      r = (Node *)addConstantNode(new ConstantNode(e->numv())) ;
   } else if (e->boundp()) { // for a bound node, index into the input nodes.
      if (e->localp())
         r = (Node *)Is[e->scopeIndex()] ;
      else if (Close)
         r = (Node *)Close[e->scopeIndex()] ;
      else {
         error("A closure-bound variable referencing outside encountered.") ;
         r = (Node *)addConstantNode(new ConstantNode(7734)) ;
      }
   } else {
      error("I didn't understand the type of node you returned.") ;
      r = (Node *)addConstantNode(new ConstantNode(7734)) ;
   }
   return r ;
}

//  To add a node by type, we have the following functions.  Each is quite
//  trivial.

void Place::draw(int x, int y) {
   int w = width() ;
   int h = height() ;
   set_color(3) ;
   draw_circle(x-w/2, y-h/2, w, h) ;
}

void Transition::draw(int x, int y) {
   int w = width() ;
   int h = height() ;
   set_color(3) ;
   draw_box(x-w/2, y-h/2, w, h) ;
}

void IfTransition::draw(int x, int y) {
   int w = width() ;
   int h = height() ;
   set_color(3) ;
   draw_box(x-w/2, y-h/2, w, h) ;
}

void ConstantNode::draw(int x, int y) {
   int w = width() ;
   int h = height() ;
   set_color(2) ;
   draw_box(x-w/2, y-h/2, w, h) ;
   label(x, y, name()) ;
}

void InputNode::draw(int x, int y) {
   int w = width() ;
   int h = height() ;
   set_color(2) ;
   draw_box(x-w/2, y-h/2, w, h) ;
   label(x, y, name()) ;
}

void PrimitiveNode::draw(int x, int y) {
   int w = width() ;
   int h = height() ;
   set_color(2) ;
   draw_box(x-w/2, y-h/2, w, h) ;
   label(x, y, name()) ;
}

void SubroutineNode::draw(int x, int y) {
   int w = width() ;
   int h = height() ;
   set_color(2) ;
   draw_box(x-w/2, y-h/2, w, h) ;
   label(x, y, name()) ;
}

void LoopNode::draw(int x, int y) {
   int w = width() ;
   int h = height() ;
   set_color(2) ;
   draw_box(x, y, w, h) ;
}

void IfNode::draw(int x, int y) {
   int w = width() ;
   int h = height() ;
   set_color(2) ;
   draw_box(x-w/2, y-h/2, w, h) ;
   label(x, y, name()) ;
}

void PetriEdge::draw(int x1, int y1, int x2, int y2) {
   set_color(3) ;
   draw_line(x1, y1, x2, y2) ;
}

void DataflowEdge::draw(int x1, int y1, int x2, int y2) {
   set_color(2) ;
   draw_line(x1, y1, x2, y2) ;
}

Place *HWSG::addPlace(Place *a) {
   places.append((elem)a) ;
   return (Place *)addNode((Node *)a) ;
}

Transition *HWSG::addTransition(Transition *a) {
   transitions.append((elem)a) ;
   return (Transition *)addNode((Node *)a) ;
}

IfTransition *HWSG::addIfTransition(IfTransition *a) {
   iftransitions.append((elem)a) ;
   return (IfTransition *)addNode((Node *)a) ;
}

ConstantNode *HWSG::addConstantNode(ConstantNode *a) {
   constantnodes.append((elem)a) ;
   return (ConstantNode *)addNode((Node *)a) ;
}

InputNode *HWSG::addInputNode(InputNode *a) {
   inputnodes.append((elem)a) ;
   return (InputNode *)addNode((Node *)a) ;
}

PrimitiveNode *HWSG::addPrimitiveNode(PrimitiveNode *a) {
   primitivenodes.append((elem)a) ;
   return (PrimitiveNode *)addNode((Node *)a) ;
}

SubroutineNode *HWSG::addSubroutineNode(SubroutineNode *a) {
   subroutinenodes.append((elem)a) ;
   return (SubroutineNode *)addNode((Node *)a) ;
}

LoopNode *HWSG::addLoopNode(LoopNode *a) {
   loopnodes.append((elem)a) ;
   return (LoopNode *)addNode((Node *)a) ;
}

IfNode *HWSG::addIfNode(IfNode *a) {
   ifnodes.append((elem)a) ;
   return (IfNode *)addNode((Node *)a) ;
}

//  To add a dataflow edge, we call the following subroutine.  We insure
//  that the head node isn't |0|, as that indicates a void value.
//  If this happens, we complain and don't hook anything up.

DataflowEdge *HWSG::addDFEdge(Node *t, Node *h, int tw =0, int hw =0) {
   if (t == 0) {
/* error("void value into something expecting a value . . . might be ok") ; */
      return 0 ;
   } else {
      DataflowEdge *e = new DataflowEdge(t, h, tw, hw) ;
      addEdge((Edge *)e) ;
      dataflowedges.append((elem)e) ;
      return e ;
   }
}

//  The petri edge routine is almost identical.

PetriEdge *HWSG::addPetriEdge(Node *t, Node *h, int tw =0, int hw =0) {
   if (t == 0) {
/* error("no way to get here . . . might be ok") ; */
      return 0 ;
   } else {
      PetriEdge *e = new PetriEdge(t, h, tw, hw) ;
      addEdge((Edge *)e) ;
      petriedges.append((elem)e) ;
      return e ;
   }
}

//  We provide a name function for those for whom it is nontrivial.

const char *ConstantNode::name() {
   char *s = tempstring() ;
   sprintf(s, "%d", val) ;
   return s ;
}

//  This function is for debug{}ing.  It simply prints out each node,
//  followed by the successors.
//
//  There's an invisible kludge in the above paragraph.  Since
//  |debug{}ging| is a preprocessor macro, and some C++ compilers
//  process comments with the preprocessor, I stick an empty pair
//  of braces in the middle of the word, which \TeX\ is smart
//  enough to mostly ignore.

void HWSG::display() {
   GraphNodeIterator gei(*this) ;
   Node* n = 0 ;
   Edge *e ;
   int i, first ;

   while (n=gei()) {
      printf("%s-%d: ", n->name(), n->serial()) ;
      for (i=0; i<n->outArity(); i++) {
         printf("(") ;
         ListIterator li(n->outEdges(i)) ;
         first = 1 ;
         while (e = (Edge *)(li())) {
            if (first)
               first = 0 ;
            else
               printf(",") ;
            printf("%s-%d<%d>", e->head()->name(), e->head()->serial(),
                                                          e->outPos()) ;
         }
         printf(")") ;
      }
      printf("\n") ;
   }
}

//  Finally, the declaration of our global |HWSG| for when we are building
//  up a graph.

HWSG *hwsg ;

//  This function is useful for building the graph of a procedure.  It 
//  allocates an `empty' graph.  Then, it takes the input nodes for the
//  procedure and allocates new nodes for them, building up an expression
//  to be evaluated.  Then, we simply call evaluate.  When we return, we
//  return the new |HWSG| and clear out the global pointer.

//  This function finds the node appropriate for a particular expression.
//  If the node has already been created, it is returned; else, a new
//  constant node is created.

Node *HWSG::exprNode(Expr *e) {
   if (e->nodep())
      return e->exprNode() ;
   else if (e->nump())
      return addConstantNode(new ConstantNode(e->numv())) ;
   else {
      experror("Needed number in graph building!", e) ;
      return addConstantNode(new ConstantNode(7734)) ;
   }
}

//  This function adds a call expression to the current graph.

Node *HWSG::callPrimGraph(Expr *fun, Expr **args, int numargs) {
   Node *callnode = 
     (Node *)addPrimitiveNode(new PrimitiveNode(fun->scopeEntry(), numargs)) ;
   int i ;
   for (i=0; i<numargs; i++, args++)
      addDFEdge(exprNode(*args), callnode, 0, i) ;
   return callnode ;
}

//  Add an if node.

Node *HWSG::addIfGraph(Expr *test, Expr *tr, Expr *fa) {
   Node *ifnode = (Node *)addIfNode(new IfNode()) ;
   addDFEdge(exprNode(test), ifnode, 0, 0) ;
   addDFEdge(exprNode(tr), ifnode, 0, 1) ;
   addDFEdge(exprNode(fa), ifnode, 0, 2) ;
   return ifnode ;
}

//  This function starts a graph out with input nodes for a lambda
//  expression.  It returns an expression to `evaluate' to figure the
//  actual graph.

Expr *HWSG::startLambda(Expr *p) {
   int inarity = p->lambdaArity() ;
   int i ;
   Expr *res = Null ;
   Expr *t ;
   Expr *arglist = p->lambdaArgs() ;
   who = p->copy() ;
   for (i=0; i<inarity; i++, arglist = arglist->cdr()) {
      Expr *n = new Expr((Node *)addInputNode(new InputNode(
                                 arglist->car()->boundScopeEntry()))) ;
      if (res == Null) {
         t = res = new Expr(n, res) ;
      } else {
         t->setcdr(new Expr(n, Null)) ;
         t = t->cdr() ;
      }
   }
   return new Expr(p, res) ;
}
