//  This file implements our Hardware Sculpture Graph, or HWSG.
//  This graph is a combination of Petri net places and transitions
//  and acyclic data-flow graph entities extended with if, subroutine,
//  and loop constructs.  It's a subclass of |Graph|.

#include "Graph.h"
#include "Scope.h"
#include "display.h"

//  First we'll discuss nodes.  Our nodes fall into the following
//  hierarchy:\:
//     \t Petri net nodes\:
//     \t\t Places\:
//     \t\t Transitions\:
//     \t\t\t Normal transitions\:
//     \t\t\t IF transitions\:
//     \t Dataflow nodes\:
//     \t\t Constants\:
//     \t\t Primitive operations\:
//     \t\t Subroutine calls\:
//     \t\t Recursive (loop) calls\:
//     \t\t IF nodes
//
//  The |Graph| class implements everything common to all of these nodes.
//  We subclass |Node| for each nonetheless, so we can implement virtual
//  functions galore.
//
//  Places in the Petri net represent some amount of state, such
//  as `wire $a$ holds value $b$', where $a$ is mentioned
//  explicitly and $b$ is a data-flow edge.

class Place : public Node {
public:
   Place(VOID) : () {}
   ~Place(VOID) {}
   virtual const char *name(VOID) { return "Place" ; }
   virtual const char *isa(VOID) { return "Place" ; }
   virtual void draw(int x, int y) ;
   virtual int width(VOID) { return MAX_NODE_HEIGHT ; }
} ;

//  Transitions in the Petri net are state transitions, where the
//  change in state is the difference between the places preceeding
//  the transition and the places succeeding the transition.

class Transition : public Node {
public:
   Transition(VOID) : () {}
   ~Transition(VOID) {}
   virtual const char *name(VOID) { return "Transition" ; }
   virtual const char *isa(VOID) { return "Transition" ; }
   virtual void draw(int x, int y) ;
   virtual int height(VOID) { return 1 ; }
   virtual int width(VOID) { return DEFAULT_NODE_WIDTH ; }
} ;

//  To manage more complex processes, we also use conditional, or
//  `if', transitions.  These take a single input data-flow edge,
//  and when the transition fires, chooses one of two sets of output
//  places to place tokens in.

class IfTransition : public Node {
public:
   IfTransition(VOID) : (2, 2) {}
   ~IfTransition(VOID) {}
   virtual const char *name(VOID) { return "IfTransition" ; }
   virtual const char *isa(VOID) { return "IfTransition" ; }
   virtual void draw(int x, int y) ;
   virtual int height(VOID) { return 1 ; }
   virtual int width(VOID) { return DEFAULT_NODE_WIDTH ; }
} ;

//  A constant data-flow node is simply a primitive operation with
//  an input arity of 0.
//  data-flow graph of the same input and output arities.

class ConstantNode : public Node {
   int val ; // we have a value to our constant nodes.
public:
   ConstantNode(int n) : () { val = n ; }
   ~ConstantNode(VOID) {}
   virtual const char *name(VOID) ;
   virtual const char *isa(VOID) { return "ConstantNode" ; }
   virtual void draw(int x, int y) ;
} ;

//  An input data-flow node corresponds to the inputs of lambda expressions.

class InputNode : public Node {
   class ScopeEntry *se ; // we need to keep track of it for later reference
public:
   InputNode(class ScopeEntry *s) : () { se = s ; }
   ~InputNode(VOID) {}
   const char *name(VOID) { return se->id() ; }
   virtual const char *isa(VOID) { return "InputNode" ; }
   virtual void draw(int x, int y) ;
} ;

//  Primitive operations can take any number
//  of inputs and generate any number of outputs.

class PrimitiveNode : public Node {
   ScopeEntry *se ; // keep this info around
public:
   PrimitiveNode(ScopeEntry *s, int arity) : (arity, 1) { se = s ; }
   ~PrimitiveNode(VOID) {}
   const char *name(VOID) { return se->id() ; }
   virtual const char *isa(VOID) { return "PrimitiveNode" ; }
   virtual void draw(int x, int y) ;
} ;

//  Subroutine calls
//  are similar, but can be expanded, and contain a pointer to a

class SubroutineNode : public Node {
   ScopeEntry *se ; // keep this info around
public:
   SubroutineNode(ScopeEntry *s, int arity) : (arity, 1) { se = s ; }
   ~SubroutineNode(VOID) {}
   const char *name(VOID) { return se->id() ; }
   virtual const char *isa(VOID) { return "SubroutineNode" ; }
   virtual void draw(int x, int y) ;
} ;

//  Loops are handled in the data-flow domain with recursive calls.
//  This could get messy when there is a lot of static context, but
//  we'll figure a way around that.

class LoopNode : public Node {
   char *nam ; // we add a name to a node
public:
   LoopNode(char *n) : () { nam = n ; }
   ~LoopNode(VOID) {}
   const char *name(VOID) { return nam ; }
   virtual const char *isa(VOID) { return "LoopNode" ; }
   virtual void draw(int x, int y) ;
} ;

//  Finally, an `if' node simply takes three inputs and generates a
//  single output.  (If multiple outputs are desired, multiple if
//  nodes can be generated.)

class IfNode : public Node {
public:
   IfNode() : (3, 1) { }
   ~IfNode(VOID) {}
   const char *name(VOID) { return "IF" ; }
   virtual const char *isa(VOID) { return "IfNode" ; }
   virtual void draw(int x, int y) ;
} ;

//  The edges we deal with fall into two groups:  Petri net edges, and
//  data-flow edges.  The former always have one end connected to a
//  |Place| and the other connected to either a |Transition| or an
//  |IfTransition|.  The latter connect a data-flow node to another
//  data-flow node, or a data-flow node to or from one or more |Place|s.
//  (These details still need refining.)

class PetriEdge : public Edge {
public:
   PetriEdge(Node *t, Node *h, int tw =0, int hw =0) : (t, h, tw, hw) {}
   virtual const char *isa(VOID) { return "PetriEdge" ; }
   virtual void draw(int x1, int y1, int x2, int y2) ;
} ;

class DataflowEdge : public Edge {
public:
   DataflowEdge(Node *t, Node *h, int tw, int hw) : (t, h, tw, hw) {}
   virtual const char *isa(VOID) { return "DataflowEdge" ; }
   virtual void draw(int x1, int y1, int x2, int y2) ;
} ;

//  Our graph structure must maintain individual lists of the |Place|s,
//  |Transition|s, |IfTransition|s, etc., for all the individual node
//  types, in addition to the master |Node| list.  We also maintain a
//  list of |PetriEdge|s as well as |DataflowEdge|s.

class HWSG : public Graph {
   List places, transitions, iftransitions, constantnodes,
        inputnodes, primitivenodes, subroutinenodes, loopnodes, ifnodes ;
   List petriedges, dataflowedges ;
   Node *outputNode ;
   Expr *who ;  // who (what expression) are we?  Must be a lambda!
public:
   HWSG(VOID) ;
   HWSG(Expr *e) ;
   ~HWSG(VOID) ;
   Node *addNode(Node *a) { return Graph::addNode(a) ; }
   Place *addPlace(Place *a) ;
   Transition *addTransition(Transition *a) ;
   IfTransition *addIfTransition(IfTransition *a) ;
   ConstantNode *addConstantNode(ConstantNode *a) ;
   InputNode *addInputNode(InputNode *a) ;
   PrimitiveNode *addPrimitiveNode(PrimitiveNode *a) ;
   SubroutineNode *addSubroutineNode(SubroutineNode *a) ;
   LoopNode *addLoopNode(LoopNode *a) ;
   IfNode *addIfNode(IfNode *a) ;
   Edge *addEdge(Edge *e) { return Graph::addEdge(e) ; }
   DataflowEdge *addDFEdge(Node *t, Node *h, int tw =0, int hw =0) ;
   PetriEdge *addPetriEdge(Node *t, Node *h, int tw =0, int hw =0) ;
   void deleteEdge(Edge *e) { Graph::deleteEdge(e) ; }
   int deleteNode(Node *n) { return Graph::deleteNode(n) ; }

//  To add an expression to this graph, we use the following function.
//  It returns a pointer to the output dataflow node.  (Needs changing.)

   Node *addExpr(Expr *e, Node **Is, Node **Close, Place **place) ;
   Node *exprNode(Expr *e) ;
   Node *callPrimGraph(Expr *fun, Expr **args, int numargs) ;
   Node *addIfGraph(Expr *test, Expr *tr, Expr *fa) ;
   Expr *startLambda(Expr *p) ;
   void display(VOID) ;
   virtual const char *isa(VOID) { return "HWSG" ; }
} ;

//  When we build up a graph, this global variable holds the actual
//  graph we construct.

extern HWSG *hwsg ;
