//  All of our functions for the |hash| class are defined in this file.

#include "Hash.h"
#include "misc.h"

//  We start with the constructor.  It takes a single argument indicating
//  the number of buckets to allocate---this should preferably be a prime
//  number.  It defaults to 7.  This function allocates and clears the
//  hash table and sets the appropriate values.

Hash::Hash(int numbuckets =7) {
   int i ;

   if (numbuckets < 5 || numbuckets > 100000)
      error("! illegal value in Hash::Hash") ;
   hashPrime = numbuckets ;
   hashtable = new struct _hashlink *[numbuckets] ;
   for (i=0; i<numbuckets; i++)
      hashtable[i] = 0 ;
   count = 0 ;
}

//  Our destructor simply deallocates what was allocated above.

Hash::~Hash() {
   int i ;
   struct _hashlink *hl, *n ;

   for (i=0; i<hashPrime; i++)
      for (hl = hashtable[i]; hl; hl = n) {
         n = hl->next ;
         delete hl ;
      }
   delete hashtable ;
}

//  We take this opportunity to define our hash function.  We use the
//  exclusive or trick.  Rather than check for negative and negate at
//  the end, which has a chance of failure that everyone seems to ignore,
//  we simply strip off the top two bits (assuming at least a 32-bit
//  long.)

int Hash::hashVal(const char *s) {
   long v = 19 ;

   while (*s)
      v = (v << 1) ^ *s++ ;
   return (v & 0x3fffffff) % hashPrime ;
}

//  Our find routine is the classic bucket hash function.

elem Hash::find(const char *s) {
   int h = hashVal(s) ;
   struct _hashlink *hl ;

   for (hl = hashtable[h]; hl; hl=hl->next)
      if (strcmp(hl->name, s)==0)
         return(hl->e) ;
   return(0) ;
}

//  This routine functions the same as the above routine, but we
//  return a pointer to the name field in the hash table itself.
//  Thus, if |s| exists in the hash table, this simply returns
//  a pointer to another copy of |s|, but this other copy is
//  permanent and unique.

char *Hash::nameFind(const char *s) {
   int h = hashVal(s) ;
   struct _hashlink *hl ;

   for (hl = hashtable[h]; hl; hl=hl->next)
      if (strcmp(hl->name, s)==0)
         return(hl->name) ;
   return(0) ;
}

//  Our add routine is also simple.  If the name already exists in the
//  hash table, we return the old value for possible deletion.

elem Hash::add(const char *s, elem el =0) {
   int h = hashVal(s) ;
   struct _hashlink *hl ;
   elem e = 0 ;

   for (hl = hashtable[h]; hl; hl = hl->next) {
      if (strcmp(hl->name, s)==0)
         break ;
   }
   if (hl == 0) {
      count++ ;
      hl = (struct _hashlink *)new char[sizeof(struct _hashlink)+strlen(s)] ;
      hl->next = hashtable[h] ;
      hashtable[h] = hl ;
      strcpy(hl->name, s) ;
   } else
      e = hl->e ;
   hl->e = el ;
   if (count > 2 * hashPrime)
      rehash() ;
   return(e) ;
}

//  The remove routine is just slightly more tricky; it has to maintain
//  a `previous' pointer so our singly-linked list can be patched.
//
//  We return |0| if the string cannot be found, or the elem if it can
//  be and was deleted.

elem Hash::remove(const char *s) {
   int h = hashVal(s) ;
   struct _hashlink *hl, *hp = 0 ;
   elem r = 0 ;

   for (hl = hashtable[h]; hl; hp = hl, hl = hl->next)
      if (strcmp(hl->name, s)==0)
         break ;
   if (hl == 0)
      return(0) ;
   if (hp == 0)
      hashtable[h] = hl->next ;
   else
      hp->next = hl->next ;
   r = hl->e ;
   delete[sizeof(struct _hashlink) + strlen(s)] (char *)hl ;
   count-- ;
   return(r) ;
}

//  Finally, if the hash table gets too full, we |rehash| to make our
//  searches quicker.  This involves finding a larger prime to use as
//  our |hashPrime|, allocating a brand new hash table, and moving
//  all of the elements.

void Hash::rehash() {
   int newPrime = hashPrime * 3 + 2 ;
   int i, h ;
   struct _hashlink *t, *p ;
   while (1) {
      for (i=3; i*i <= newPrime; i+=2)
         if (newPrime % i == 0)
            break ;
      if (i*i > newPrime)
         break ;
      else
         newPrime += 2 ;
   }
   i = newPrime ; newPrime = hashPrime ; hashPrime = i ;
   struct _hashlink **newhashtable = new struct _hashlink *[hashPrime] ;
   for (i=0; i<hashPrime; i++)
      newhashtable[i] = 0 ;
   for (i=0; i<newPrime; i++)
      for (t=hashtable[i]; t; t=p) {
         p = t->next ;
         h = hashVal(t->name) ;
         t->next = newhashtable[h] ;
         newhashtable[h] = t ;
      }
   delete hashtable ;
   hashtable = newhashtable ;
}
