//  The implementation of the integer set capability.

#include "IntSet.h"

//  Creates a new integer set, using the little buffer if possible.

IntSet::IntSet(int nsize) {
   size = nsize ;
   wsize = (size + 31) >> 5 ;
   if (wsize <= 1) {
      bits = &forshortsets ;
      forshortsets = 0 ;
   } else {
      bits = new int [wsize] ;
      clear() ;
   }
}

//  Clears an integer set.

void IntSet::clear(VOID) {
   int i ;
   for (i=0; i<wsize; i++)
      bits[i] = 0 ;
}

//  Hashes the integer set with respect to a particular hash modulus.

int IntSet::hash(int mod) {
   int i, s ;
   for (i=0, s=0; i<wsize; i++)
      if (bits[i] < 0)
         s += (~bits[i]) % mod ;
      else
         s += bits[i] % mod ;
   return s ;
}

//  Checks for set equivalence.

int IntSet::same(IntSet *p) {
   int i ;
   if (size == p->size) {
      for (i=0; i<wsize; i++)
         if (bits[i] != p->bits[i])
            return 0 ;
      return 1 ;
   } else
      return 0 ;
}

//  Copies an integer set.

IntSet *IntSet::copy(VOID) {
   IntSet *ret = new IntSet(size) ;
   int i ;
   for (i=0; i<wsize; i++)
      ret->bits[i] = bits[i] ;
   return ret ;
}

//  This function returns the smallest member of the set.  If there is no
//  such, it returns -1.
//
//  I had a clever algorithm, described below, but as it turns out, it's
//  slower than a simpler algorithm.  So we use the simpler algorithm.
//  So much for being clever with Fermat's little theorem.
//
//  There's a trick to this.  Specifically, we use
//  the fact that $2^n\pmod 37$ is unique for the set $(0\ldots 31)$.  We
//  thus use a simple least-bit extraction trick in conjunction with a
//  modulus and table lookup.  The table |magic| is such that
//  $|magic[|2^{n+1}\pmod 37-1|]|=n$.  Note that $k\hbox{xor}(k-1)=2^{n+1}-1$,
//  where $n$ is the bit number of the least significant bit set in the
//  word.  Due to the stupidity of the modulus operator in C, our elegant
//  algorithm has to be marred by a sign test.

static char magic[] = {
  -1,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
   5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
   6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
   5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
   7,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
   5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
   6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
   5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0 } ;

int IntSet::smallestMember(VOID) {
   int i, j, s ;
   for (i=0, s=0; i<wsize; i++, s += 32)
      if (j=bits[i])
         if (j & 0xffff)
            if (j & 0xff)
               return s + magic[j & 0xff] ;
            else
               return s + 8 + magic[(j >> 8) & 0xff] ;
         else
            if (j & 0xff0000)
               return s + 16 + magic[(j >> 16) & 0xff] ;
            else
               return s + 24 + magic[(j >> 24) & 0xff] ;
   return -1 ;
}
