//  This file defines a simple list class.  The operations we wish to
//  perform on this list include insertion at any point, deletion at any
//  point, and a general traversal forwards or backwards.  We might also
//  want to know the number of elements in the list at a given point.

#ifndef LIST_CLASS
#define LIST_CLASS
#include "Object.h"

//  We wish to use this class for both homogenous and hetergenous
//  lists.  In addition, we are only concerned with the list at this
//  point; we do not wish to add any artificial constraints on the
//  number of times a particular object can occur in a list.  Thus,
//  our actual list `element' is simply a |void *| that can be
//  coerced to any appropriate type as required.  This is actually
//  defined in |misc.h|.
//
//  To implement such a beast, we use three classes.  One, |link|,
//  holds the list links, backwards and forwards, and our element.
//  The second is the list structure itself, and the third is our
//  general iterator class.  These are all friends.
//
//  Most of the work is actually done inside the link class, maintaining
//  the links backwards and forwards.  Both |insert| and |append| take
//  |link *| parameters rather than |elem| parameters so we can later
//  implement list reversal and other reorderings efficiently and easily.

class link {
   friend class List ;
   friend class ListIterator ;
   link *next, *prev ;
   elem e ;
public:
   void remove(VOID) { next->prev = prev; prev->next = next; delete this; }
   void removeNoDealloc(VOID) { next->prev = prev ; prev->next = next ; }
   void insert(link *what) ;
   void append(link *what) ;
   link(elem a) { e=a ; }
   link(VOID) {}
   elem element(VOID) { return e ; }
   void setLink(elem a) { e=a ; }
} ;

//  Our list class builds on the above link class, adding a count field.
//  The |head| link is useful for easily representing an empty list, while
//  making it convenient to find the first and last element.  In addition,
//  information particular to a list can be put onto the head |elem|
//  field if so desired.
//
//  Most of these functions deal with |link|s, because there is nothing
//  that says the elements of the list need be unique, so we cannot simply
//  traverse the list looking for something to delete.
//
//  The destructor is somewhat clever; it simply removes the first
//  element until everything, even the |head|, is gone.

class List : public Object {
   friend class ListIterator ;
   link* head ;
   int count ;
public:
   elem insert(elem a) { head->append(new link(a)) ; count++; return(a); }
   elem append(elem a) { head->insert(new link(a)) ; count++; return(a); }
   void insert(link *a) { head->append(a) ; count++ ; }
   void append(link *a) { head->insert(a) ; count++ ; }
   link *first(VOID) ;
   elem firstEl(VOID) { return el(first()) ; }
   elem takeFirstEl(VOID) ;
   link *last(VOID) ;
   elem lastEl(VOID) { return el(last()) ; }
   link *next(link *what) ;
   link *prev(link *what) ;
   int length(VOID) { return count ; }
   elem el(link *what) ;
   elem& headel(VOID) { return (head->e) ; }
   void remove(link *what) { what->remove() ; count-- ; }
   void removeNoDealloc(link *what) { what->removeNoDealloc() ; count-- ; }
   link *find(elem e) ;          // looks for elem with simple pointer compare
   int findAndRemove(elem e) ;   // returns 1 if found, 0 if not
   List(elem hdata =0) ;
   ~List(VOID) ;
   virtual const char *isa(VOID) { return "List" ; }
} ;

//  The last class we need is our general iteration class.  This is
//  where we maintain the `current' link so we can easily delete,
//  insert, etc.  Our default operator |()| simply returns the next
//  element until there aren't any more.  Then it returns |0|, so
//  no list can have |0| as one of its elements.
//
//  Note that when the list is done (ie, |next()| or |prev()| return
//  |0|), the list is `reprimed', so calling the same function again
//  will return the first (or last) element of the list.

class ListIterator {
   List *ourlist ;
   link *curplace ;
public:
   ListIterator(List *s) { ourlist = s ; curplace = s->head ; }
   elem operator()(VOID) ;
   link *curlink(VOID) { return curplace ; }
   elem next(VOID) ;
   elem prev(VOID) ;
   elem removeForward(VOID) ;   // remove the currelem elem, moving forward
   elem removeBackward(VOID) ;  // remove the current elem, moving back
   void reset(VOID) { curplace = ourlist->head ; }
} ;

//  To make it easy to delete a list of things, we provide the following
//  macro.  Note that this does *not* delete the list itself, just the
//  objects themselves.

#define deleteAListOf(type,list) { \
   class ListIterator li(list); type *e ; while (e=(type *)li()) delete e ; }

//  We also want to make it easy to define a particular type of list
//  iterator.

#define defineListIterator(name, type, itype, initializor) \
class name : ListIterator { \
public: \
   name(itype) : (initializor) {} \
   type operator()(VOID) { return (type)(ListIterator::operator()()); } \
   type next(VOID) { return (type)(ListIterator::next()); } \
   type prev(VOID) { return (type)(ListIterator::prev()); } \
   type removeForward(VOID) { return (type)(ListIterator::removeForward()); } \
   type removeBackward(VOID) \
                           { return (type)(ListIterator::removeBackward()); } \
} ;
#endif
