//  This file contains all of the very primitive functions in our system.

#include "HWSG.h"
#include "Key.h"

//  Every so often we need to grab the global scope.

extern Scope *scope ;

//  We use a define to check the number of arguments.

static char BADARGS[] = "Wrong number of arguments to primitive function" ;
#define checkargs(a,b) {if((a)!=(b))\
                           {error(BADARGS);return(Null);}}

//  A function to check that all arguments are numerical.

int checknumbers(Expr **a, int n) {
   while (n > 0) {
      if (!((*a)->nump())) {
         experror("expected only numerical arguments to arithmetic function",
                          *a) ;
         return(1) ;
      }
      n-- ; a++ ;
   }
   return 0 ;
}

//  Actually, before we get into that we define a `default' function that
//  just returns |Null| no matter what it is passed.  We also use this as
//  our |void| routine, to evaluate something but throw the result away.

Expr *nullreturn(Expr **, int) {
   return Null ;
}

//  Now the standard set of three Lisp functions.

Expr *Cons(Expr **a, int n) {
   checkargs(2, n) ;
   return new Expr(a[0]->copy(), a[1]->copy()) ;
}

Expr *Car(Expr **a, int n) {
   checkargs(1, n) ;
   return a[0]->car()->copy() ;
}

Expr *Cdr(Expr **a, int n) {
   checkargs(1, n) ;
   return a[0]->cdr()->copy() ;
}

//  The |length| primitive feeds off the |length| function in |Expr|.

Expr *Length(Expr **a, int n) {
   checkargs(1, n) ;
   if (a[0]->arrayp())
      return new Expr(a[0]->arrayLength()) ;
   else if (a[0]->stringp())
      return new Expr(strlen(a[0]->string())) ;
   else
      return new Expr(a[0]->length()) ;
}

//  The |obnum| returns an integer of the object described.  This integer
//  is guaranteed to be unique.  If no argument is passed, we return the
//  total number of objects ever allocated.

Expr *ObNum(Expr **a, int n) {
   if (n > 0) {
      checkargs(1, n) ;
      return new Expr(a[0]->serial()) ;
   } else
      return new Expr(Object_sergen) ;
}

//  The |ObActive| function returns the number of objects still allocated.

Expr *ObActive(Expr **, int) {
   return new Expr(Object_sergen - Object_deleted) ;
}

//  The |list| function simply builds and returns a list of its arguments.

Expr *PrimList(Expr **a, int n) {
   Expr *ret = Null ;
   while (n > 0) {
      n-- ;
      ret = new Expr(a[n]->copy(), ret) ;
   }
   return ret ;
}

//  The |append| function is somewhat more complex.  We can use a
//  simple reference copy of the last argument, but all the others
//  must be constructed fresh.  This looks quite complicated, and
//  that's because we're building the list from the beginning down
//  rather than the end up.

Expr *Append(Expr **e, int n) {
   Expr *ret = Null ;
   int i ;
   if (n == 0)
      return (Null) ;
   Expr *r = Null ;
   i = 0 ;
   while (i < n-1) {
      Expr *nl = *e++ ;
      while (nl != Null) {
         if (ret == Null) {
            r = ret = new Expr(nl->car()->copy(), Null) ;
         } else {
            r->setcdr(new Expr(nl->car()->copy(), Null)) ;
            r = r->cdr() ;
         }
         nl = nl->cdr() ;
      }
      i++ ;
   }
   if (ret == Null)
      return (*e)->copy() ;
   else {
      r->setcdr((*e)->copy()) ;
      return(ret) ;
   }
}

//  Are we a pair or an atom?

Expr *Pairp(Expr **a, int n) {
   checkargs(1, n) ;
   if (a[0]->consp())
      return True ;
   else
      return Null ;
}

Expr *Atomp(Expr **a, int n) {
   checkargs(1, n) ;
   if (a[0]->consp())
      return Null ;
   else
      return True ;
}

Expr *StringP(Expr **a, int n) {
   checkargs(1, n) ;
   if (a[0]->stringp())
      return True ;
   else
      return Null ;
}

Expr *NumP(Expr **a, int n) {
   checkargs(1, n) ;
   if (a[0]->nump())
      return True ;
   else
      return Null ;
}

//  Some simple mathematical operators.  The |+| operator takes
//  any number of numerical operands and sums them.

Expr *Plus(Expr **a, int n) {
   int s = 0 ;
   if (checknumbers(a, n))
      return(BadNum) ;
   while (n > 0) {
      n-- ;
      s += (*a++)->numv() ;
   }
   return new Expr(s) ;
}

//  The |-| operator subtracts all numbers after the first from
//  the first number.

Expr *Sub(Expr **a, int n) {
   if (checknumbers(a, n))
      return(BadNum) ;
   if (n <= 0)
      return(new Expr(0)) ;
   int s = (*a++)->numv() ;
   n-- ;
   while (n > 0) {
      n-- ;
      s -= (*a++)->numv() ;
   }
   return new Expr(s) ;
}

//  A quick |1-| operator.

Expr *Dec(Expr **a, int n) {
   Expr *r ;
   if (n != 1) {
      error("bad number of arguments to 1-") ;
   } else if ((r=a[0])->nump()==0) {
      error("needed a number for 1-") ;
   } else {
      return new Expr(r->numv()-1) ;
   }
   return BadNum ;
}

//  The |*| operator simply multiplies its arguments.

Expr *Times(Expr **a, int n) {
   int s = 1 ;
   if (checknumbers(a, n))
      return(BadNum) ;
   while (n > 0) {
      n-- ;
      s *= (*a++)->numv() ;
   }
   return new Expr(s) ;
}

//  The |/| operator divides all numbers after the first from
//  the first number.

Expr *Divide(Expr **a, int n) {
   if (checknumbers(a, n))
      return(BadNum) ;
   if (n <= 0)
      return(new Expr(0)) ;
   int s = (*a++)->numv() ;
   int d ;
   n-- ;
   while (n-- > 0) {
      d = (*a++)->numv() ;
      if (d==0)
         error("divide by zero error") ;
      else
         s /= d ;
   }
   return new Expr(s) ;
}

//  The modulo operator |%| expects two numeric arguments and it
//  finds the remainder when the first is divided by the
//  second.

Expr *Remainder(Expr **a, int n) {
   checkargs(2, n) ;
   if (checknumbers(a, n))
      return(BadNum) ;
   return new Expr(a[0]->numv() % a[1]->numv()) ;
}

//  The operator |odd| returns |True| if the number is
//  odd and |Null| otherwise.

Expr *Odd(Expr **a, int n) {
   checkargs(1, n) ;
   if (checknumbers(a, n))
      return(BadNum) ;
   return (a[0]->numv() & 1) ? True : Null ;
}

//  The logical and, or, exclusive or, and not functions are trivial.

Expr *Land(Expr **a, int n) {
   int res = -1 ;
   if (checknumbers(a, n))
      return(BadNum) ;
   while (n-- > 0)
      res &= (*a++)->numv() ;
   return new Expr(res) ;
}

Expr *Lor(Expr **a, int n) {
   int res = 0 ;
   if (checknumbers(a, n))
      return(BadNum) ;
   while (n-- > 0)
      res |= (*a++)->numv() ;
   return new Expr(res) ;
}

Expr *Lxor(Expr **a, int n) {
   int res = 0 ;
   if (checknumbers(a, n))
      return(BadNum) ;
   while (n-- > 0)
      res ^= (*a++)->numv() ;
   return new Expr(res) ;
}

Expr *Lnot(Expr **a, int n) {
   checkargs(1, n) ;
   if (checknumbers(a, n))
      return(BadNum) ;
   return new Expr(~(a[0]->numv())) ;
}

//  The array primitives are next.  We can set, clear, and allocate an
//  array.

Expr *ARef(Expr **a, int n) {
   Expr *ar = a[0] ;
   while (n > 1) {
      if (a[1]->nump()==0) {
         error("need a number as args to aref") ;
         return Null ;
      }
      ar = ar->aref(a[1]->numv()) ;
      n-- ;
      a++ ;
   }
   return ar->copy() ;
}

Expr *ASet(Expr **a, int n) {
   Expr *ar = a[0] ;
   if (n < 3) {
      error("too few args to aset") ;
      return Null ;
   }
   while (n > 3) {
      if (a[1]->nump()==0) {
         error("need a number as args to aref") ;
         return Null ;
      }
      ar = ar->aref(a[1]->numv()) ;
      n-- ;
      a++ ;
   }
   ar->aset(a[1]->numv(), a[2]->copy()) ;
   return Null ;
}

Expr *recurseMakeArray(Expr **a, int n) {
   Expr *ar ;
   int i ;

   ar = a[0]->makeArray(a[0]->numv()) ;
   if (n > 1) {
      for (i=a[0]->numv()-1; i >= 0; i--)
         ar->aset(i, recurseMakeArray(a+1, n-1)) ;
   }
   return ar ;
}

Expr *MakeArray(Expr **a, int n) {
   if (n < 1) {
      error("too few args to makearray") ;
      return Null ;
   }
   if (a[0] == True)
      return MakeArray(a+1, n-1)->structArray() ;
   if (checknumbers(a, n))
      return(BadNum) ;
   return recurseMakeArray(a, n) ;
}

//  The operator |random| returns a random number from 0 to one less
//  than the passed value.

Expr *Random(Expr **a, int n) {
   extern unsigned long rand() ;
   int t = (rand() >> 2) ;
   checkargs(1, n) ;
   if (checknumbers(a, n))
      return(BadNum) ;
   return new Expr((t ^ (t >> 16)) % a[0]->numv()) ;
}

//  The numerical equals |=| expects two numeric arguments and it
//  returns |Null| if they are not equal or |#t| if they are.

Expr *NumEqual(Expr **a, int n) {
   checkargs(2, n) ;
   if (checknumbers(a, n))
      return(BadNum) ;
   return ((a[0]->numv() == a[1]->numv()) ? True : Null) ;
}

//  The numerical not equals (|!=| or |<>|) expects two numeric arguments
//  and it returns |Null| if they are equal or |#t| if they are not.

Expr *NotEqual(Expr **a, int n) {
   checkargs(2, n) ;
   if (checknumbers(a, n))
      return(BadNum) ;
   return ((a[0]->numv() != a[1]->numv()) ? True : Null) ;
}

//  Some more comparators.

Expr *LessThan(Expr **a, int n) {
   int t, ne ;

   if (checknumbers(a, n))
      return(BadNum) ;
   if (n < 2)
      return True ;
   t = (*a++)->numv() ;
   n-- ;
   while (n > 0) {
      ne = (*a++)->numv() ;
      if (t >= ne)
         return(Null) ;
      else
         t = ne ;
      n-- ;
   }
   return True ;
}

//  Another comparator.

Expr *LessThanEqual(Expr **a, int n) {
   int t, ne ;

   if (checknumbers(a, n))
      return(BadNum) ;
   if (n < 2)
      return True ;
   t = (*a++)->numv() ;
   n-- ;
   while (n > 0) {
      ne = (*a++)->numv() ;
      if (t > ne)
         return(Null) ;
      else
         t = ne ;
      n-- ;
   }
   return True ;
}

//  Yet another.

Expr *GreaterThan(Expr **a, int n) {
   int t, ne ;

   if (checknumbers(a, n))
      return(BadNum) ;
   if (n < 2)
      return True ;
   t = (*a++)->numv() ;
   n-- ;
   while (n > 0) {
      ne = (*a++)->numv() ;
      if (t <= ne)
         return(Null) ;
      else
         t = ne ;
      n-- ;
   }
   return True ;
}

//  And the last.

Expr *GreaterThanEqual(Expr **a, int n) {
   int t, ne ;

   if (checknumbers(a, n))
      return(BadNum) ;
   if (n < 2)
      return True ;
   t = (*a++)->numv() ;
   n-- ;
   while (n > 0) {
      ne = (*a++)->numv() ;
      if (t < ne)
         return(Null) ;
      else
         t = ne ;
      n-- ;
   }
   return True ;
}

//  The |not| operator returns |True| if its single parameter is |Null|
//  or |True| otherwise.

Expr *Not(Expr **a, int n) {
   checkargs(1, n) ;
   return ((a[0] == Null) ? True : Null) ;
}

//  The |eq| operator returns |True| if its two parameters are the same
//  expression.

Expr *Eq(Expr **a, int n) {
   checkargs(2, n) ;
   return ((a[0]==a[1]) ? True : Null) ;
}

//  Turns an integer into a string with the specified radix.

static char *digs = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ" ;

Expr *IntStr(Expr **a, int n) {
   int radix ;
   unsigned int v ;
   char *p ;
   int negative = 0 ;
   char tempstr[40] ;

   if (checknumbers(a, n))
      return BadNum ;
   if (n > 2 || n < 1) {
      error("need one or two args to num->str") ;
      return Null ;
   }
   if (n > 1)
      radix = a[1]->numv() ;
   else
      radix = 10 ;
   if (radix < 2 || radix > 36) {
      error("radix out of range in num->str") ;
      radix = 10 ;
   }
   p = tempstr + 38 ;
   *p-- = 0 ;
   v = a[0]->numv() ;
   if ((v & 0x80000000)) {
      v = (~v) + 1 ;
      negative = 1 ;
   }
   while (1) {
      *p-- = digs[v % radix] ;
      v /= radix ;
      if (v==0)
         break ;
   }
   if (negative)
      *p-- = '-' ;
   return new Expr(p + 1, EXPR_STRING) ;
}

//  Concatenates a bunch of things to form a string.

Expr *StrCat(Expr **a, int n) {
   char tempstr[400] ;
   char *p = tempstr ;

   while (n-- > 0) {
      if ((*a)->nump())
         sprintf(p, "%d", (*a)->numv()) ;
      else if ((*a)->stringp())
         strcpy(p, (*a)->string()) ;
      else if ((*a)->idp())
         strcpy(p, (*a)->id()) ;
      else
         strcpy(p, "<?>") ;
      a++ ;
      p += strlen(p) ;
   }
   return new Expr((const char *)(tempstr), EXPR_STRING) ;
}

//  Compares two strings.

Expr *StrCmp(Expr **a, int n) {
   checkargs(n, 2) ;
   if (a[0]->stringp() && a[1]->stringp())
      return new Expr(strcmp(a[0]->string(), a[1]->string())) ;
   else {
      error("only expected strings in strcmp function") ;
      return Null ;
   }
}

//  The |value| operator turns a wire into a number.  This is the only way
//  to observe the value of a wire.

Expr *Value(Expr **a, int n) {
   checkargs(n, 1) ;
   Expr *e = a[0] ;
   if (e->wirep()) {
      return new Expr(e->wireValue()) ;
   } else {
      error("Can't take the `value' of a non-wire") ;
      return Null ;
   }
}

//  Our |clear| function calls the global |clearscope()| routine.

extern void clearscope() ;

Expr *Clear(Expr **, int) {
   clearscope() ;
   return Null ;
}

//  To set or clear the debug flags, we use the following routine.

Expr *Debug(Expr **a, int n) {
   checkargs(1, n) ;
   Expr *e = a[0] ;
   if (e->stringp()) {
      set_debug_flags(e->string()) ;
      return True ;
   } else if (e->nump()) {
      debug_flags = e->numv() ;
      return True ;
   } else
      return Null ;
}

//  The |display| function just calls the appropriate member for all of its
//  arguments.

Expr *Display(Expr **a, int n) {
   int pos = 0 ;
   while (n > 0) {
      pos = (*a++)->display(stdout, 0, pos) ;
      n-- ;
   }
   fflush(stdout) ;
   return Null ;
}

//  Some file handling functions.

Expr *FOpen(Expr **a, int n) {
   checkargs(2, n) ;
   Expr *e = a[0] ;
   Expr *m = a[1] ;
   FILE *f ;
   if (e->stringp() && m->stringp() && (f = fopen(e->string(), m->string())))
      return new Expr(f) ;
   else
      return Null ;
}

Expr *FClose(Expr **a, int n) {
   checkargs(1, n) ;
   Expr *e = a[0] ;
   if (e->filep())
      e->closefile() ;
   return True ;
}

//  The |load| function expects a string.  It attempts to open this name as
//  a file for input, and then it calls the main |readevalprint| loop over
//  this file.  It returns |#t| if it could open the file; |Null| otherwise.
//  Only if a second argument is supplied is output written to standard output.

extern void readevalprint(FILE *, FILE *) ;
extern void filedisplay(const char *) ;

Expr *Load(Expr **a, int n) {
   Expr *e = a[0] ;
   if (e->stringp()) {
      FILE *f = fopen(e->string(), "r") ;
      if (f==0) {
         error2("Could not open %s for loading", e->string()) ;
         return(Null) ;
      } else {
         filedisplay(e->string()) ;
         readevalprint(f, ((n > 1) ? stdout : 0)) ;
         fclose(f) ;
         return(True) ;
      }
   } else {
      error("expected a string argument to load") ;
      return(Null) ;
   }
}

//  The |System| call invokes an external Unix command and returns
//  the error code.

#ifdef AMIGA
void Execute(const char *, int, int) ;
#else
#ifndef GPLUSPLUS
int system(const char *) ;
#endif
#endif

Expr *System(Expr **a, int n) {
   Expr *e = a[0] ;
   checkargs(n, 1) ;
   if (e->stringp()) {
#ifdef AMIGA
      Execute(e->string(), 0, 0) ;
      return new Expr(0) ;
#else
      return new Expr(system(e->string())) ;
#endif
   } else {
      error("expected a string argument to system") ;
      return(Null) ;
   }
}

Expr *StrToSym(Expr **a, int n) {
   checkargs(n, 1) ;
   if (a[0]->stringp())
      return new Expr(a[0]->string()) ;
   else
      return Null ;
}

//  If we need to exit, we have this simple routine.

Expr *Exit(Expr **, int) {
   printf("User exit requested\n") ;
   clearscope() ;
   exit(0) ;
   return Null ;
}

//  Creates a graph.

Expr *GraphExpr(Expr **a, int n) {
   Expr *e = a[0] ;
   checkargs(n, 1) ;
   return new Expr(new HWSG(e)) ;
}

//  Displays a graph.

Expr *DisplayGraph(Expr **a, int n) {
   Expr *e = a[0] ;
   int created = 0 ;
   checkargs(n, 1) ;
   if (e->lambdap()) {
      e = new Expr(new HWSG(e)) ;
      created = 1 ;
   }
   if (e->graphp()) {
      e->exprGraph()->display() ;
      if (created)
         e->remove() ;
      return True ;
   } else {
      return Null ;
   }
}

//  Creates a window.

Expr *MakeWindow(Expr **a, int n) {
   Expr *e = a[0] ;
   int created = 0 ;
   Expr *r = Null ;
   checkargs(n, 1) ;
   if (e->lambdap()) {
      e = new Expr(new HWSG(e)) ;
      created = 1 ;
   }
   if (e->graphp()) {
      cur_display = &S_display ;
      r = e->exprGraph()->makeWindow(e) ;
      if (created)
         e->remove() ;
   }
   return r ;
}

//  Writes a PostScript file.

extern FILE *psfile ;

Expr *PSGen(Expr **a, int n) {
   Expr *e = a[0] ;
   Expr *f = a[1] ;
   int created = 0 ;
   Expr *r = Null ;
   checkargs(n, 2) ;
   if (f->stringp() == 0 ||
       (psfile = fopen(f->string(), "w"))==0) {
      return r ;
   }
   if (e->lambdap()) {
      e = new Expr(new HWSG(e)) ;
      created = 1 ;
   }
   if (e->graphp()) {
      cur_display = &P_display ;
      r = e->exprGraph()->makeWindow(e) ;
      if (created)
         e->remove() ;
      r->remove() ;
      r = True ;
   }
   return r ;
}

//  To make a key, we call this.

Expr *MakeKey(Expr **, int) {
   return new Expr(new Key) ;
}

//  Some things to do manipulation of graphs by hand.

Expr *NewGraph(Expr **, int) {
   return new Expr(new HWSG()) ;
}

Expr *AddNode(Expr **a, int n) {
   checkargs(n, 2) ;
   if (a[0]->graphp()==0 || a[1]->stringp()==0) {
      error("bad type in AddNode") ;
      return Null ;
   }
   a[0]->exprGraph()->addNamedNode(a[1]->string()) ;
   return Null ;
}

Expr *AddEdge(Expr **a, int n) {
   checkargs(n, 3) ;
   if (a[0]->graphp()==0 || a[1]->stringp()==0 || a[2]->stringp()==0) {
      error("bad type in AddEdge") ;
      return Null ;
   }
   a[0]->exprGraph()->addNamedEdge(a[1]->string(), a[2]->string()) ;
   return Null ;
}

Expr *DeleteNode(Expr **a, int n) {
   checkargs(n, 2) ;
   if (a[0]->graphp()==0 || a[1]->stringp()==0) {
      error("bad type in DeleteNode") ;
      return Null ;
   }
   a[0]->exprGraph()->deleteNamedNode(a[1]->string()) ;
   return Null ;
}

Expr *DeleteEdge(Expr **a, int n) {
   checkargs(n, 3) ;
   if (a[0]->graphp()==0 || a[1]->stringp()==0 || a[2]->stringp()==0) {
      error("bad type in DeleteEdge") ;
      return Null ;
   }
   a[0]->exprGraph()->deleteNamedEdge(a[1]->string(), a[2]->string()) ;
   return Null ;
}

//  This is called at the beginning to register all of our primitives in a
//  particular (usually global) scope.  If you hide these with your own
//  definitions, you die.

extern void initforms(Scope *scope) ; // find this in |EvalState.cc|

void initprimitives(Scope *scope) {
   evalftype *f ;
#ifdef AMIGA
//   inittimer() ;
#endif
   scope->add("void", new Expr(f=nullreturn)) ;
   scope->add("cons", new Expr(f=Cons)) ;
   scope->add("car", new Expr(f=Car)) ;
   scope->add("cdr", new Expr(f=Cdr)) ;
   scope->add("length", new Expr(f=Length)) ;
   scope->add("obnum", new Expr(f=ObNum)) ;
   scope->add("obactive", new Expr(f=ObActive)) ;
   scope->add("list", new Expr(f=PrimList)) ;
   scope->add("append", new Expr(f=Append)) ;
   scope->add("pair?", new Expr(f=Pairp)) ;
   scope->add("atom?", new Expr(f=Atomp)) ;
   scope->add("string?", new Expr(f=StringP)) ;
   scope->add("num?", new Expr(f=NumP)) ;
   scope->add("+", new Expr(f=Plus, HARDPRIM)) ;
   scope->add("-", new Expr(f=Sub, HARDPRIM)) ;
   scope->add("1-", new Expr(f=Dec, HARDPRIM)) ;
   scope->add("*", new Expr(f=Times, HARDPRIM)) ;
   scope->add("/", new Expr(f=Divide, HARDPRIM)) ;
   scope->add("%", new Expr(f=Remainder, HARDPRIM)) ;
   scope->add("odd", new Expr(f=Odd, HARDPRIM)) ;
   scope->add("land", new Expr(f=Land, HARDPRIM)) ;
   scope->add("lor", new Expr(f=Lor, HARDPRIM)) ;
   scope->add("lxor", new Expr(f=Lxor, HARDPRIM)) ;
   scope->add("lnot", new Expr(f=Lnot, HARDPRIM)) ;
   scope->add("aref", new Expr(f=ARef)) ;
   scope->add("aset", new Expr(f=ASet)) ;
   scope->add("make-array", new Expr(f=MakeArray)) ;
   scope->add("random", new Expr(f=Random, HARDPRIM)) ;
   scope->add("=", new Expr(f=NumEqual, HARDPRIM)) ;
   scope->add("<>", new Expr(f=NotEqual, HARDPRIM)) ;
   scope->add("!=", new Expr(f=NotEqual, HARDPRIM)) ;
   scope->add("<", new Expr(f=LessThan, HARDPRIM)) ;
   scope->add("<=", new Expr(f=LessThanEqual, HARDPRIM)) ;
   scope->add(">", new Expr(f=GreaterThan, HARDPRIM)) ;
   scope->add(">=", new Expr(f=GreaterThanEqual, HARDPRIM)) ;
   scope->add("not", new Expr(f=Not, HARDPRIM)) ;
   scope->add("eq", new Expr(f=Eq, HARDPRIM)) ;
   scope->add("value", new Expr(f=Value, HARDPRIM)) ;
   scope->add("clear", new Expr(f=Clear)) ;
   scope->add("debug", new Expr(f=Debug)) ;
   scope->add("load", new Expr(f=Load)) ;
   scope->add("system", new Expr(f=System)) ;
   scope->add("exit", new Expr(f=Exit)) ;
   scope->add("display", new Expr(f=Display)) ;
   scope->add("open", new Expr(f=FOpen)) ;
   scope->add("close", new Expr(f=FClose)) ;
   scope->add("graph", new Expr(f=GraphExpr)) ;
   scope->add("display-graph", new Expr(f=DisplayGraph)) ;
   scope->add("window", new Expr(f=MakeWindow)) ;
   scope->add("ps-to-file", new Expr(f=PSGen)) ;
   scope->add("make-key", new Expr(f=MakeKey)) ;
   scope->add("newgraph", new Expr(f=NewGraph)) ;
   scope->add("addedge", new Expr(f=AddEdge)) ;
   scope->add("deleteedge", new Expr(f=DeleteEdge)) ;
   scope->add("addnode", new Expr(f=AddNode)) ;
   scope->add("deletenode", new Expr(f=DeleteNode)) ;
   scope->add("num->str", new Expr(f=IntStr)) ;
   scope->add("strcat", new Expr(f=StrCat)) ;
   scope->add("strcmp", new Expr(f=StrCmp)) ;
   scope->add("str->sym", new Expr(f=StrToSym)) ;
   scope->add("#t", True->forever()) ;      // we permanent |True| here
   scope->add("#f", Null->forever()) ;
   scope->add("#u", Undefined->forever()) ;
   scope->add("#?", BadNum->forever()) ;
   initforms(scope) ;
}
