//  This file implements some common algorithms for STGs, based on simple
//  graphs.

#include "Graph.h"
#include "IntSet.h"

//  This function searches forward through a graph for a path from one
//  node to another.  If one is found, it is marked and true is returned;
//  if none is found, none is returned.

int Graph::searchForwardLessMisc(Node *from, Node *to, Edge *toignore) {
   int i ;
   if (from == to)
      return 1 ;
   for (i=0; i<from->outArity(); i++) {
      ListIterator li(from->outEdges(i)) ;
      Edge *e ;
      while (e = (Edge *)(li())) {
         if (e != toignore && e->tail()->misc == 0) {
            e->tail()->misc = 1 ;
            if (searchForwardLessMisc(e->tail(), to, e))
               return 1 ;
            e->tail()->misc = 0 ;
         }
      }
   }
   return 0 ;
}

//  This first function looks for redundant edges in a graph.  A redundant
//  edge is one that is subsumed by other edges; for sample, |A->B| is
//  redundant if there are edges |A->C| and |C->B|.

Edge *Graph::redundantEdge(VOID) {
   GraphNodeIterator gni(*this) ;
   Node *n ;
   while (n=gni())
      n->misc = 0 ;
   GraphEdgeIterator gei(*this) ;
   Edge *e ;
   while (e=gei()) {
      Node *tail = e->tail() ;
      Node *head = e->head() ;
      if (searchForwardLessMisc(head, tail, e))
         return e ;
   }
   return 0 ;
}

//  This function checks to make sure that each node has at least one
//  input and one output edge.

Node *Graph::IOConnected(VOID) {
   GraphNodeIterator gni(*this) ;
   Node *n ;
   while (n=gni())
      if (n->inCount() <= 0 || n->outCount() <= 0)
         return n ;
   return 0 ;
}

//  This function searches forward through a graph for a path from one
//  node to another and removes all edges on such a path from the passed
//  int set.  It returns `1' if any such paths were found.

int Graph::removeAllPathsFrom(Node *from, Node *to, class IntSet *is) {
   int i ;
   int ret = 0 ;
   if (from == to)
      return 1 ;
   for (i=0; i<from->outArity(); i++) {
      ListIterator li(from->outEdges(i)) ;
      Edge *e ;
      while (e = (Edge *)(li())) {
         if (e->tail()->misc == 0) {
            e->tail()->misc = 1 ;
            if (removeAllPathsFrom(e->tail(), to, is)) {
               ret = 1 ;
               is->subInt(e->misc) ;
            }
            e->tail()->misc = 0 ;
         }
      }
   }
   return ret ;
}

//  This function constructs the reachability graph for the STG.  It
//  first finds an initial marking that satisfies one token per loop
//  and a token in every loop.  

#define REACHPRIME (1001)
struct reachStruct {
   struct reachStruct *next ; /* for the hash list */
   IntSet *intset ;
   struct Node *node ;
} ;

struct reachStruct *Graph::interpret(IntSet *marking,
       struct reachStruct **hashTable, Graph *underConstruction,
       int *tokencount) {
   int h = marking->hash(REACHPRIME) ;
   struct reachStruct *ret ;
   for (ret = hashTable[h]; ret; ret = ret->next)
      if (marking->same(ret->intset))
         break ;
   if (ret)
      return ret ;
   ret = new reachStruct ;
   ret->next = hashTable[h] ;
   hashTable[h] = ret ;
   ret->intset = marking->copy() ;
   ret->node = underConstruction->createNode() ;
   GraphNodeIterator gni(*this) ;
   Node *n ;
   while (n=gni()) {
      Edge *e ;
      if (tokencount[n->misc] == 0) { // it's enabled
         {  ListIterator li(n->inEdges(0)) ;
            while (e = (Edge *)(li())) {
               marking->subInt(e->misc) ;
               tokencount[n->misc]++ ;
            }
         }
         {  ListIterator li(n->outEdges(0)) ;
            while (e = (Edge *)(li())) {
               marking->addInt(e->misc) ;
               tokencount[e->head()->misc]-- ;
            }
         }
         underConstruction->createEdge(ret->node, interpret(marking,
                           hashTable, underConstruction, tokencount)->node) ;
         {  ListIterator li(n->inEdges(0)) ;
            while (e = (Edge *)(li())) {
               marking->addInt(e->misc) ;
               tokencount[n->misc]-- ;
            }
         }
         {  ListIterator li(n->outEdges(0)) ;
            while (e = (Edge *)(li())) {
               marking->subInt(e->misc) ;
               tokencount[e->head()->misc]++ ;
            }
         }
      }
   }
   return ret ;
}

Graph *Graph::reachabilityGraph(VOID) {
   int i, numplaces, numtransitions ;
   GraphNodeIterator gni(*this) ;
   GraphEdgeIterator gei(*this) ;
   Node *n ;
   Edge *e ;
   IntSet *curmark, *nextmark ;

//  First, we count the number of places and transitions and number the
//  edges.

   numplaces = edgeList()->length() ;
   numtransitions = nodeList()->length() ;
   Edge **edges = new Edge *[numplaces] ;
   for (i=0; e=gei(); i++) {
      e->misc = i ;
      edges[i] = e ;
   }

//  Next, we find some initial marking.  We do this by putting all the
//  edges in the set curmark.  Then, we extract an edge from curmark,
//  and make it a marked edge by adding it to nextmark.  Finally, we
//  find all loops that include the edge and remove them from nextmark.
//  (These are the loops that have at least one token in them already.)
//  This algorithm ensures that each loop has exactly one token in it.

   curmark = new IntSet(numplaces) ;
   for (i=0; i<numplaces; i++)
      curmark->addInt(i) ;
   nextmark = new IntSet(numplaces) ;
   for (i=0; n=gni(); i++)
      n->misc = 0 ;
   while (1) {
      i = curmark->smallestMember() ;
      if (i < 0)
         break ;
      nextmark->addInt(i) ;
      curmark->subInt(i) ;
      e = edges[i] ;
      removeAllPathsFrom(e->head(), e->tail(), curmark) ;
   }

//  At this point we have an initial marking.  We want to recursively
//  interpret the Petri net, stopping whenever we encounter a marking
//  we've seen before, and explore all possible markings.  As we do so,
//  we want to build up an explicit reachability graph with a node for
//  each possible marking.

   int *tokens = new int [numplaces] ;
   Graph *newg = new Graph() ;
   struct reachStruct **rhash = new struct reachStruct *[REACHPRIME] ;
   for (i=0; i<REACHPRIME; i++)
      rhash[i] = 0 ;
   for (i=0, gni.reset(); n=gni(); i++) {
      n->misc = i ;
      tokens[i] = n->inCount() ;
   }
   (void)interpret(nextmark, rhash, newg, tokens) ;
   delete edges ;
   delete curmark ;
   delete nextmark ;
   delete tokens ;
   delete rhash ;
   return newg ;
}
