//  This file implements the `Type' class, that holds the types for
//  all of our variables.  The language is not strictly strongly typed
//  at the moment, but this should change.

#include "Scope.h"

//  Our first task is the contrustor that builds a type from an
//  expression.  If the expression is not a list that starts with
//  the keyword |type|, we return |UNSPECIFIED|.  Otherwise, we
//  scan the list, looking for `keywords'.
//
//  A type specification has the following form:

 //   <type> := <id>                      |
 //             ( type <id> <keyword>* )  |
 //             ( type <id> ( <type>* ) )
 //   <id> := id | ()
 //   <keyword> := <integer> | wire | value | noreturn | key | unsigned

//  The first form of |type| is an |UNSPECIFIED| type.  The second
//  is a `normal' wire or value.  The third is a closure type,
//  where the list of types corresponds to the type it returns and
//  the types of arguments it takes.

Type::Type(Expr *e) {
   const char *myid ;
   int i ;
   Expr *t, *tt ;

   closuretypes = 0 ;
   id = 0 ;
   flags = 0 ;
   width = 1 ;
   type = TYPE_UNSPECIFIED ;
   if (e->consp() == 0) { // not a list; return |UNSPECIFIED|
      type = TYPE_UNSPECIFIED ;
      myid = e->searchId() ;
      if (myid)
         id = newstring(myid) ;
   } else if ((myid = e->carid()) && strcmp(myid, "type")==0) {
      t = e->cdr()->car() ;
      if (t!=Null) {
         myid = t->searchId() ;
         if (myid == 0)
            error("expected identifier or () in type") ;
         else {
            id = newstring(myid) ;
         }
      }
      t = e->cdr()->cdr() ;
      if (t->car()->consp()) { // closure
         t = t->car() ;
         width = t->length() ;
         if (width < 1)
            error("closure types must return something") ;
         closuretypes = new Type *[width] ;
         for (i=0; i<width; i++, t=t->cdr())
            closuretypes[i] = new Type(t->car()) ;
         type = TYPE_CLOSURE ;
      } else {                 // just keywords
         for (; t != Null; t = t->cdr()) {
            tt = t->car() ;
            if (tt->nump()) {
               width = tt->numv() ;
               if (type == TYPE_UNSPECIFIED)
                  type = TYPE_VALUE ;
            } else if (tt->idp()) {
               myid = tt->id() ;
               if (strcmp(myid, "wire")==0) {
                  type = TYPE_WIRE ;
               } else if (strcmp(myid, "key")==0) {
                  type = TYPE_KEY ;
               } else if (strcmp(myid, "value")==0) {
                  type = TYPE_VALUE ;
               } else if (strcmp(myid, "noreturn")==0) {
                  type = TYPE_NORETURN ;
               } else if (strcmp(myid, "unsigned")==0 ||
                                           strcmp(myid, "u")==0) {
                  flags |= TYPEF_UNSIGNED ;
               } else if (strcmp(myid, "in")==0) {
                  flags |= TYPEF_INPUT_ONLY ;
               } else if (strcmp(myid, "out")==0) {
                  flags |= TYPEF_OUTPUT_ONLY ;
               } else {
                  error2("illegal keyword `%s' ignored in type specification",
                                                                       myid) ;
               }
            } else {
               error("illegal expression ignored in type specification") ;
            }
         }
      }
   } else {
      e->display(stderr, 0, 0) ;
      error(" bad form in type specification; returning |UNSPECIFIED|.") ;
      type = TYPE_UNSPECIFIED ;
   }
   if (closuretypes == 0) {
      mask = (1L << width) - 1 ;
      if ((flags & TYPEF_UNSIGNED) == 0)
         smask = 1L << (width - 1) ;
      else
         smask = 0 ;
   }
   if (debugging('t'))
      this->display() ;
}

//  Our destructor is straightforward.

Type::~Type() {
   int i ;
   if (type == TYPE_CLOSURE) {
      for (i=0; i<width; i++)
         delete closuretypes[i] ;
      delete closuretypes ;
   }
   if (id) {
      delete id ;
      id = 0 ;
   }
}

//  Fixes a value according to the masks, regardless of the type (so check
//  the type before you use this function.)

int Type::fixVal(int n) {
   n &= mask ;
   if (n & smask)
      n |= ~mask ; // turn on all the bits above the mask
   return n ;
}

//  We need to cast something to a particular type.  This routine does the
//  magic.

Expr *Type::typeCast(Expr *e) {
   int newval ;

   if (this == 0 || type == TYPE_UNSPECIFIED)
      return e ; // no type information; just return expr
   switch(type) {
case TYPE_VALUE :
      if (e->nump()) {
         newval = fixVal(e->numv()) ;
         if (newval == e->numv()) // minor optimization
            return e ;
         else {
            e->remove() ;
            return new Expr(newval) ;
         }
      } else if (e->nodep()) {
         return e ;
      } else {
         error("type error:  expected integer") ;
         return e ;
      }

//  For wires, the width must most definitely match.

case TYPE_WIRE :
      if (e->wirep()) {
         if (e->wireWidth() != width)
            error("type error:  mixing wires of different widths") ;
         return e ;
      } else {
         error("type error:  expected wire") ;
         return e ;
      }
case TYPE_CLOSURE :
      if (e->procp())
         return e ; // if it's a closure type, just check that we have a proc
      else {
         error("type error:  expected closure type") ;
         return e ;
      }
case TYPE_KEY :
      if (e->keyp())
         return e ;
      else {
         error("type error:  expected key type") ;
         return e ;
      }
case TYPE_NORETURN :
      error("returning into a noreturn type???") ;
      return e ;
case TYPE_UNSPECIFIED :
      return e ;
default:
      error("! internal error; invalid type type") ;
      return e ;
   }
}

//  For debugg{}ing, we have a display routine.

static char *typenames[] = { "Value", "Wire", "Closure", "NoReturn",
   "Key", "Unspecified" } ;

void Type::display(int level =0) {
   int i ;
   for (i=0; i<level; i++)
      printf(" ") ;
   if (this == 0)
      printf("Type = 0 ... internal or something\n") ;
   else if (type < 0 || type > TYPE_UNSPECIFIED)
      printf("Invalid type!!! (%ld) internal error . . .\n", this) ;
   else {
      printf("Type `%s' = %s[%d] ", (id ? id : "*NULL*"), typenames[type],
                                                                       width) ;
      if (flags & TYPEF_UNSIGNED)
         printf("(unsigned) ") ;
      if (flags & TYPEF_INPUT_ONLY)
         printf("(input) ") ;
      if (flags & TYPEF_OUTPUT_ONLY)
         printf("(output) ") ;
      printf("\n") ;
   }
   if (type == TYPE_CLOSURE)
      for (i=0; i<width; i++)
         closuretypes[i]->display(level+3) ;
}
