/*
/    This very simple program `weaves' some code into a nice, clean,
/    \TeX\ file for printing.  Comments are processed by \TeX\ and code is
/    printed verbatim.  It's a far cry from a real web, since it doesn't
/    perform any macro expansion or indexing, but it's better than line
/    printer output.
/
/    This code should work with virtually any language, if the code is
/    written with this program in mind.
/
/    Our primary strategy is this.  The very first character of the file
/    is considered the `comment' character (almost all files start with
/    a comment.)  This and any subsequent lines starting with that
/    character are considered a comment, and handed straight through to
/    \TeX, after stripping the comment character, any non-space characters
/    following, and all following consecutive spaces.  (This way we can
/    handle the |/*| comments from standard C, for instance.)
/    Anything else is bracketed by |\\beginverb{`\\\\}| and
/    |\\endverb|, except that any single backslashes are turned into
/    double backslashes so TeX doesn't get confused.
/
/    Comments that do not start at the very left-most margin are ignored.
/
/    There are some files we wish to include completely in \TeX\ mode.
/    An example of these is a documentation file, where we wish to illustrate
/    rather than execute the macros.  There are also some files we wish to
/    set entirely in comment mode, such as a Makefile.  The former is indicated
/    by the second character of the file being a |%|, the latter by it being
/    an exclamation point.  This is ugly and arbitrary perhaps, but easily
/    changed.
/*/

#include "stdio.h"

/*
/   We allow lines up to MAXLINELENGTH chars.  Our |getline| procedure is
/   just like standard C's, except we strip off any spaces at the end of
/   the line.
/*/

#define MAXLINELENGTH (255)
char line[MAXLINELENGTH] ;
int getline() {
   char *p = line ;
   int c ;

   while (1) {
      c = getchar() ;
      if ((c==EOF) || (c==10)) {
         *p = 0 ;
         while (p > line && *(p-1) <= ' ')
            *(--p) = 0 ;
         return(c == 10 || p > line) ;
      }
      *p++ = c ;
   }
}

/*
/   We call the |output()| subroutine to write a line of output.  The
/   |outputsub()| subroutine does the same thing, only replacing any
/   backslashes with a doubled backslash.
/*/

output(s)
char *s ;
{
   while (*s) {
      putchar(*s) ;
      s++ ;
   }
   putchar(10) ;
}

outputsub(s)
char *s ;
{
   while (*s) {
      putchar(*s) ;
      if (*s == '\\')
         putchar(*s) ;
      s++ ;
   }
   putchar(10) ;
}

/*
/   Finally, our main routine.  We need to maintain a bit of state; are we
/   in a commented section?  Have we seen any blank lines?  We also count
/   the number of characters and lines in the file; at a later date we
/   may choose to make some use of this information.
/*/

main(argc, argv)
int argc ;
char *argv[] ;
{
   int incode = 0 ;
   int pendinglines = 0 ;
   char *p ;
   int commchar ;
   long numchars, numlines ;
   long tnumchars = 0, tnumlines = 0 ;

   output("\\input cpw.mac") ;  /* read in the auxillary macros */
   while (argc > 1) {       /* process all files */
      argc-- ;
      argv++ ;
      if (freopen(argv[0], "r", stdin)==NULL) {
         fprintf(stderr, "Couldn't open %s\n", argv[0]) ;
         continue ;
      }
      commchar = -1 ;
      sprintf(line, "\\begfile{%s}", argv[0]) ;
      output(line) ;
      numchars = 0 ;
      numlines = 0 ;
      while (getline()) {

/*
/   We are at the first line of a file if |commchar| is -1.  In this case,
/   we stash the first character as the comment character.  If the second
/   character is an exclamation point, we set the |commchar| to 0, which
/   means match anything, and thus the entire file will be done in \TeX.
/   If the second character is a percent sign, we set |commchar| to -2,
/   which will never match anything.
/
/   Note the way we `cache' blank lines; this makes a blank line not switch
/   us from one mode to another.
/*/

         if (commchar == -1) {
            commchar = line[0] ;
            if (line[1] == '!')
               commchar = -2 ;
            else if (line[1] == '%')
               commchar = 0 ;
         }
         numlines++ ;
         numchars += strlen(line) ;
         if (line[0] == 0)
            pendinglines++ ;
         else {

/*
/   If the current line is a comment line, then we need to go to \TeX\ mode.
/   If we are in code mode, we switch to \TeX\ mode, and then output all
/   pending lines (\TeX\ will not collapse pending lines in code mode.)
/   Next, we skip over the rest of the word and the intervening space, and
/   output the line, literally.
/*/

            if (commchar == 0 || line[0] == commchar) {
               if (incode) {
                  output("\\endverb") ;
                  incode = 0 ;
               }
               while (pendinglines) {
                  output("") ;
                  pendinglines-- ;
               }
               p = line + 1 ;
               while (*p && *p > ' ')  /* skip the rest of the word */
                  p++ ;
               while (*p && *p <= ' ') /* skip the blank space */
                  p++ ;
               output(p) ;
/*
/   If the current line is not a comment, then we need to be in code
/   mode.  We flush all pending lines and start code mode.
/*/

            } else {
               while (pendinglines) {
                  output("") ;
                  pendinglines-- ;
               }
               if (! incode) {
                  output("\\begverb{`\\\\}") ;
                  incode = 1 ;
               }
               outputsub(line) ;
            }
         }
      }

/*
/   We finish a file, and then the entire job, with this code.
/*/

      tnumlines += numlines ;
      tnumchars += numchars ;
      if (incode) {
         output("\\endverb") ;
         incode = 0 ;
      }
      sprintf(line, "\\endfile{%s}{%ld}{%ld}", argv[0], numlines, numchars) ;
      output(line) ;
   }
   sprintf(line, "\\done{%ld}{%ld}", tnumlines, tnumchars) ;
   output(line) ;
   exit(0) ;
}
