//  This is our main program.

#include "Scope.h"
#include "stdio.h"

//  We need a global scope and a current scope.

Scope globalscope, *scope ;

//  This routine reads a single token from an input stream.
//  A token is either `(', `)', `.', or a sequence of alphanumerics.
//  This routine also skips comments (anything between a `;' and a
//  newline.)

#define MAXTOKLENGTH (200)
static char thistoken[MAXTOKLENGTH+3] ;
static char attributes[256] ; // the attributes of a char

//  Sets the character attributes.

void setup_attributes(VOID) {
   attributes['('] = 1 ; // start list
   attributes[')'] = 1 ; // end list
   attributes['.'] = 1 ; // dotted pair
   attributes['\''] = 2 ; // quote
   attributes[','] = 2 ; // comma (unquote)
   attributes['`'] = 2 ; // quasiquote
}

char *gettoken(FILE *f) {
   int c ;
   char *p ;

//  If the last token hasn't been used, we return it.

   if (thistoken[0])
      return(thistoken) ;

//  The first thing we do when searching for a new token is to skip past
//  any whitespace or comments that might get in our way.

   while (1) {
      c = getc(f) ;
      if (c == ';') {
         while (1) {
            c = getc(f) ;
            if (c == 10 || c == EOF)
               break ;
         }
      }
      if (c == EOF)
         return(0) ;
      if (c > ' ')
         break ;
   }
   p = thistoken ;
   *p++ = c ;

//  If the first character is a double quote, then we have a string.
//  We collect everything until the next double quote.  What we return
//  will not have the closing double quote, so we don't need to delete
//  it.
//
//  We use a C-style syntax, converting backslash.  We support the
//  newline, backslash, double-quote, return, and octal character
//  escapes.

   if (c == '"') {
      while (1) {
         c = getc(f) ;
         if (c=='"' || c==EOF || p > thistoken + MAXTOKLENGTH) {
            *p = 0 ;
            break ;
         } else if (c=='\\') {
            c = getc(f) ;
            if (c=='n')
               *p++ = 10 ;
            else if (c=='r')
               *p++ = 13 ;
            else if (c=='b')
               *p++ = 8 ;
            else if (c=='e')
               *p++ = 27 ;
            else if ('0' <= c && c <= 7) {
               c -= '0' ;
               c = (c << 3) + (getc(f) - '0') ;
               c = (c << 3) + (getc(f) - '0') ;
               *p++ = c ;
            } else
               *p++ = c ;
         } else
            *p++ = c ;
      }
      return(thistoken) ;
   }

//  Next we consider the case of individual character tokens.  The
//  only ones we currently recogize are `.', '(', ')', and `''.

   if (attributes[c]) {
      *p = 0 ;
      return(thistoken) ;
   }

//  Otherwise we snarf up characters until we run out of room.
//  Or we hit a character <= ' ', or equal to '(', ')', '.'.
//  In any case, we `ungetc' that character for the next time around.

   while (1) {
      c = getc(f) ;
      if (c > ' ' && attributes[c] == 0) {
         *p++ = c ;
         if (p > thistoken + MAXTOKLENGTH) {
            error("input token too long; truncated") ;
            while (1) {
               c = getc(f) ;
               if (c <= ' ' || attributes[c])
                  break ;
            }
            break ;
         }
      } else
         break ;
   }
   *p = 0 ;
   if (c != EOF)
      ungetc(c, f) ;
   return(thistoken) ;
}

//  To go to the next token, call this.

#define usetoken(f) { thistoken[0] = 0 ; }

//  We need a subroutine to scan hex numbers or decimal numbers.

long myatol(const char *s) {
   int negative = 0 ;
   int d ;
   long v = 0 ;

   if (*s == '-') {
      negative = 1 ;
      s++ ;
   }
   if (s[1] == 'x' && *s == '0') { // hex number
      s += 2 ;
      while (d = *s++) {
         if ('A' <= d && d <= 'F')
            d -= ('A' - 10) ;
         else if ('a' <= d && d <= 'f')
            d -= ('a' - 10) ;
         else if ('0' <= d && d <= '9')
            d -= '0' ;
         else
            d = 0 ;
         v = (v << 4) + d ;
      }
   } else { // just call atol
      v = atol(s) ;
   }
   return negative ? -v : v ;
}

//  This subroutine takes an id of the form |foo:bar:baz:boom| and converts
//  it to an |Expr| of the form |(type foo bar baz boom)|.  We might
//  need to recursively call |readexpr| if there is a list pending, since
//  a closure type is represented as the list of types of its arguments.

Expr *readexpr(FILE *f) ;

Expr *maketypelist(char *s, FILE *f) {
   char *p = s ;
   Expr *top = new Expr(Literal_type, Null) ;
   Expr *cur = top ;
   Expr *t ;
   int addtail = 0 ;
   while (*p) {
      int numsonly = 1 ;
      while (*s && *s != ':') {
         if ('0' > *s || *s > '9')
            numsonly = 0 ;
         s++ ;
      }
      if (*s == ':') {
         *s++ = 0 ;
         if (*s == 0)
            addtail = 1 ;
      }
      if (numsonly) {
         if (*p==0)
            t = new Expr(Null, Null) ;
         else
            t = new Expr(new Expr((int)atol(p)), Null) ;
      } else
         t = new Expr(new Expr(p), Null) ;
      cur->setcdr(t) ;
      cur = t ;
      p = s ;
   }
   if (addtail) {
      usetoken(f) ;
      t = new Expr(readexpr(f), Null) ;
      cur->setcdr(t) ;
   }
   return(top) ;
}

//  Our first concern is reading an expression and `parsing' it.
//  We need to forward-declare our |readlist| procedure.

Expr *readlist(FILE *f) ;

Expr *readexpr(FILE *f) {
   char *s = gettoken(f) ;
   if (s == 0)
      return(0) ;

//  If the first token is an open parenthesis, we call the auxilliary
//  procedure |readlist| to get the entities, which must be suffixed
//  by a close parenthesis.

   if (*s == '(') {
      usetoken(f) ;
      Expr *l = readlist(f) ;
      s = gettoken(f) ;
      if (*s != ')')
         error("Missing parenthesis inserted") ;
      else
         usetoken(f) ;
      return(l) ;

//  Handling a `quote' is remarkably easy.  We simply get the next token
//  recursively, and then transform `a' to `|(quote |a|)|'.  The same
//  for the other special tokens.

   } else if (attributes[*s] == 2) {
      Expr *tt ;
      switch(*s) {
      case '\'': tt = Literal_quote ; break ;
      case ',': 
         int c = getc(f) ;
         if (c == '@')
            tt = Literal_unquote_splicing ;
         else {
            if (c != EOF)
               ungetc(c, f) ;
            tt = Literal_unquote ;
         }
         break ;
      case '`': tt = Literal_quasiquote ; break ;
      }
      usetoken(f) ;
      Expr *t = readexpr(f) ;
      if (t == 0)
         return tt ;
      else
         return new Expr(tt, new Expr(t, Null)) ;

//  If the first character is a double quote, we create a string and
//  return a string.


   } else if (*s == '"') {
      usetoken(f) ;
      return new Expr(s+1, EXPR_STRING) ;

//  If we fell off the end, we die.

   } else if (*s == '.' || *s == ')') {
      return(0) ;
   } else {

//  If the token is some sort of id or number, we parse it here.  First,
//  we determine if it is a number; if so, we convert it to an integer
//  and create the appropriate |Expr|.  If it has a colon in it, then it
//  is a typed variable, so we call an auxilliary routine to handle this.
//  Otherwise it's a simple |id|.

      int hasalpha = 0 ;
      int hascolon = 0 ;
      char *p = s ;
      if (*p == '-' && ('0' <= p[1] && p[1] <= '9'))
         p++ ;       // allow a leading minus sign for numbers
      if (*p == '0' && p[1] == 'x' && (('0' <= p[2] && p[2] <= '9') ||
                                       ('a' <= p[2] && p[2] <= 'f') ||
                                       ('A' <= p[2] && p[2] <= 'F')))
         p += strlen(p) ; // allow hex input for numbers; `assume' correct form
      while (*p) {
         if (*p == ':') {
            hasalpha = 1 ;
            hascolon = 1 ;
         } else if ('0' > *p || *p > '9')
            hasalpha = 1 ;
         p++ ;
      }
      if (hascolon) {
         Expr *r = maketypelist(s, f) ;
         usetoken(f) ;
         return(r) ;
      } else if (hasalpha) {
         Expr *r = new Expr(s) ;
         usetoken(f) ;
         return r ;
      } else {
         int i = myatol(s) ;
         usetoken(f) ;
         return new Expr(i) ;
      }
   }
}

//  This routine is responsible for reading a list.  It just grabs
//  expressions and tacks them on to the end of the list until it
//  hits a close parenthesis.

Expr *readlist(FILE *f) {
   Expr *b ;
   Expr *a = readexpr(f) ;
   Expr *r ;
   if (a) {
      r = new Expr(a, Null) ;
      a = r ;
      while (b = readexpr(f)) {
         b = new Expr(b, Null) ;
         a->setcdr(b) ;
         a = b ;
      }
      if (thistoken[0] == '.') {
         usetoken(f) ;
         b = readexpr(f) ;
         if (b == 0)
            error("missing expr after `.'") ;
         a->setcdr(b) ;
      }
   } else
      r = Null ;
   return(r) ;
}

//  Our top-level is a read-eval-print loop, just like in Scheme.
//  Note that we |usetoken(stdin)| at the end to flush any token we had
//  to look at.  We only need to look at a token if it is something
//  like an extra, unneeded `)' or `.'.

void readevalprint(FILE *fin, FILE *fout) {
   Expr *e = 0 ;
   int localsreq ;
   while (1) {
      if (fin == stdin)
         printf("-> ") ;
      if (fout)
         fflush(fout) ;
      localsreq = 0 ;
      e = readexpr(fin) ;
      if (e) {
         e = e->macroexpand() ;
         if (debugging('m'))
            if (e->display(stdout))
               printf("\n") ;
         e = e->alphaBind(scope, &localsreq) ;
#ifdef DEBUG
         if (localsreq)
            printf("Top level requires %d locals\n", localsreq) ;
#endif
         if (debugging('b'))
            if (e->display(stdout))
               printf("\n") ;
         e = e->eval(localsreq) ;

//  Another Lattice C bug on the Amiga; I can't separate these with |&&| like
//  I want to; I have to resort to really strange tricks.  This coercion will
//  give a warning, but somehow it will force the correct code to be generated.

         if (((long)fout) && e->display(fout))
            fprintf(fout, "\n") ;
      }
      if (thistoken[0]) {
         if (thistoken[0]==')')
            error("Extra `)' ignored") ;
         else if (thistoken[0]=='.')
            error("Extra `.' ignored") ;
         else
            error("Extra token ignored") ;
         usetoken(stdin) ;
      } else if (e == NULL)
         break ;
      if (e)
         e->remove() ;
   }
}

//  This function is called to `push' our global scope.

void pushscope() {
   scope = scope->push() ;
}

//  And this function deletes one level of scope, replacing it
//  with another.

void clearscope() {
   scope = scope->pop()->push() ;
}

void filedisplay(const char *s) {
   fprintf(stdout, "<%s>", s) ; fflush(stdout) ;
}

//  This is where we put it all together.

main(int argc, char *argv[]) {
   FILE *f ;

   startup(argc, argv) ;
   setup_attributes() ;
   scope = &globalscope ;
   initprimitives(scope) ;
   if (f=fopen("hws.hs", "r")) {
      filedisplay("hws.hs") ;
      readevalprint(f, 0) ;
      fclose(f) ;
   }
   pushscope() ;
   readevalprint(stdin, stdout) ;
   clearscope() ;
   exit(0) ;
}
