//  This file defines some standard C-style utilities for any C++
//  project we happen to build, such as error reporting and string
//  handling.

#include "misc.h"
#include "Expr.h"

//  Our generalized error routine prints out its argument.  If the first
//  character is an exclamation point, the error is fatal and we exit.

char *progname ;
void error(const char *s) {
   fprintf(stderr, "%s: %s\n", progname, s) ;
   if (*s == '!')
      exit(1) ;
}

//  The routine `error2' is nice when we want to mention some string.

void error2(const char *f, const char *s) {
   fprintf(stderr, "%s: ", progname) ;
   fprintf(stderr, f, s) ;
   fprintf(stderr, "\n") ;
   if (*f == '!')
      exit(1) ;
}

void experror(const char *s, Expr *e) {
   fprintf(stderr, ">>> ") ;
   e->display(stderr) ;
   fprintf(stderr, "\n") ;
   error(s) ;
}

//  We need to provide an error handler for out of memory since we don't
//  test our |new| calls for zero.  We simply exit when we run out of
//  memory.

void (*_new_handler)() ;

void no_memory() {
   error("! out of memory") ;
}

//  If we need a permanent copy of a string, we call this routine.  There's
//  no backing out, but this is just test code anyway.

char *newstring(const char *s) {
   char *p = new char[strlen(s)+1] ;
   strcpy(p, s) ;
   return(p) ;
}

//  This routine maintains a circular string buffer for temporary use.
//  You call it to get a pointer to a region you want to use.  You
//  had better use it before the next call, and it better be used with
//  a null-terminated string, because subsequent calls make use of that
//  null termination to find the next safe starting point.

#define TEMPBUFSIZE (8192)
static char tempbuf[TEMPBUFSIZE] ;
static char *tempptr = tempbuf ;

extern char *tempstring(void) {
   tempptr += strlen(tempptr) + 1 ;
   if (tempptr > tempbuf + TEMPBUFSIZE - 260)
      tempptr = tempbuf ;
   *tempptr = 0 ;
   return(tempptr) ;
}

//  This routine sets the debug flags from a string.  We need a variable
//  to hold these debug flags.

long debug_flags ;

void set_debug_flags(const char *p) {
   int i ;

   if (*p) {
      debug_flags = 0 ;
      while (*p) {
         i = *p++ ;
         if ('A' <= i && i <= 'Z')
            i += 32 ;
         if ('a' <= i && i <= 'z')
            debug_flags |= 1L << (i-'a') ;
      }
   } else
      debug_flags = -1 ;
}

//  For timing things, we have the |time| function.  It might be moderately
//  expensive to call and the moularity may not be perfect, so try to call
//  it on fairly slow things.  The time returned is total elapsed CPU time
//  since the beginning of the program.  The value returned is in
//  milliseconds.  But first, the system dependent parts.  (These should
//  probably be collected into a separate file some day.)

#ifdef AMIGA
extern void DateStamp(long *) ;
static long initt[3] ;
int ourtime() {
   long t[3] ;
   DateStamp(t) ;
   return(20 * (t[2]-initt[2] + 3000 * (t[1]-initt[1] +
                  1440 * (t[0]-initt[0])))) ;
}
void inittimer() {
   DateStamp(initt) ;
}
#else
#undef KERNEL
#include <sys/times.h>
#ifndef HZ
#define HZ CLK_TCK
#endif

int ourtime() {
   struct tms buf ;
   times(&buf) ;
   return((buf.tms_utime + buf.tms_stime) * 1000 / HZ) ;
}
#endif

//  We need a bit bucket to dump output we don't care about to.

FILE *stdbitbucket ;

#ifdef AMIGA
#define DEVNULL "nil:"
#else
#define DEVNULL "/dev/null"
#endif

void startup(int argc, char **argv) {
   char *p ;
   progname = argv[0] ;
#ifdef AMIGA
   inittimer() ;
#endif
   _new_handler = no_memory ;
   stdbitbucket = fopen(DEVNULL, "w") ;
   while (argc > 1) {
      argc-- ;
      argv++ ;
      p = *argv ;
      if (*p == '-' && p[1] == 'd')
         set_debug_flags(p+2) ;
   }
}
