;
;   This is a synthesis of my firing of cyclic petri nets and the analysis
;   of acyclic timed nets in hopes that it will yield something that will
;   analyze cyclic timed nets.
;
;   Load structures.
;
(load "struct.hs")
(load "quick.hs")
;
(defstruct place
   name            ; the name of this place
   type            ; one of constraint, elemental, or spring
   mintime         ; the minimum time
   maxtime         ; the maximum time
   satisfied?      ; has this constraint been met yet?
   pred            ; predecessor transitions
   succ            ; successor transitions
   token           ; one of none, normal, or initial
   )
;
(define constraint 'constraint)
(define elemental 'elemental)
(define none 'none)
(define initial 'initial)
;
(defstruct transition
   name            ; the name of this transition
   firecount       ; how many times has it fired so far
   firingtimes     ; when did it last fire, and so on in a list
   pred            ; predecessor places; indices
   succ            ; successor transition
   tokcount        ; how many tokens in predecessor places
   predcount       ; how many predecessor places
   )
;
(defstruct graph
   name
   places         ; a list of all the places
   transitions
   loopcount      ; number of times we've gone around
   constraints-left   ; number of constraints that haven't worked yet
   enabled        ; a list of enabled transitions
   next-enabled   ; a list of enabled transitions on the next go-round
   )
;
;   Allocate a global graph.
;
(define (new-tpn)
;  (new)
   (define *tpn* (new-graph))
   (set-graph-name *tpn* "global-tpn"))
;
(new-tpn)
;
;   Allocate and initialize a place.
;
(define (make-place name type mintime maxtime)
   (let ((p (new-place)))
      (set-place-name p name)
      (set-place-type p type)
      (set-place-mintime p mintime)
      (set-place-maxtime p maxtime)
      (set-place-token p 'none)
      p))
;
;   Ditto, transition.
;
(define (make-transition name)
   (let ((t (new-transition)))
      (set-transition-name t name)
      (set-transition-firecount t 0)
      (set-transition-tokcount t 0)
      (set-transition-predcount t 0)
      t))
;
(define (add-transition name)
   (let ((t (make-transition name)))
      (set-graph-transitions *tpn* (cons t (graph-transitions *tpn*)))
      t))
;
(define (add-transitions . list)
   (let loop ((list list))
      (if list
         (cons
            (add-transition (car list))
            (loop (cdr list))))))
;
(define (find-transition name)
   (let loop ((list (graph-transitions *tpn*)))
      (if list
         (if (= 0 (strcmp name (transition-name (car list))))
             (car list)
             (loop (cdr list)))
         ())))
;
(define (add-place from to type mintime maxtime)
;  (->q from to)
   (let ((p (make-place (strcat from "->" to) type mintime maxtime))
         (from (find-transition from))
         (to (find-transition to)))
      (set-graph-places *tpn* (cons p (graph-places *tpn*)))
      (set-place-pred p (cons from (place-pred p)))
      (set-place-succ p (cons to (place-succ p)))
      (set-transition-succ from (cons p (transition-succ from)))
      (set-transition-pred to (cons p (transition-pred to)))
      p))
;
(define (find-place name)
   (let loop ((list (graph-places *tpn*)))
      (if list
         (if (= 0 (strcmp name (place-name (car list))))
             (car list)
             (loop (cdr list)))
         ())))
;
(defmacro (for-all-places var cmd)
   `(dolist ,var (graph-places *tpn*) ,cmd))
;
(defmacro (for-all-transitions var cmd)
   `(dolist ,var (graph-transitions *tpn*) ,cmd))
;
(define (enabled t)
   (= (transition-tokcount t) (transition-predcount t)))
;
(define (mark-as-enabled t)
   (let ((phase (- (transition-firecount t) (graph-loopcount *tpn*))))
      (if (= phase 0)
         (set-graph-enabled *tpn*
            (cons t (graph-enabled *tpn*)))
         (if (= phase 1)
            (set-graph-next-enabled *tpn*
               (cons t (graph-next-enabled *tpn*))))
            (error "Phase error!"))))
;
;   Also, terminate after four times around.
;
(define MAXLOOPS 4)
;
(define (get-next-to-fire)
   (let ((test (graph-enabled *tpn*)))
      (if test
         (begin
            (set-graph-enabled *tpn* (cdr test))
            (car test))
         (let ((test (graph-next-enabled *tpn*)))
            (if test
               (begin
                  (display "Tick:  Graph rolled over "
                            (graph-loopcount *tpn*) "\n")
                  (set-graph-enabled *tpn* (cdr test))
                  (set-graph-next-enabled *tpn* ())
                  (set-graph-loopcount *tpn* (+ 1 (graph-loopcount *tpn*)))
                  (if (< (graph-loopcount *tpn*) MAXLOOPS)
                     (car test) ()))
               (error "No more enabled."))))))
;
(define (initial-marking . foo)
   (set-graph-constraints-left *tpn* 0)
   (for-all-places p
      (begin
         (set-place-token p none)
         (if (not (elemental? p))
            (set-graph-constraints-left *tpn*
                               (+ 1 (graph-constraints-left *tpn*))))))
   (for-all-transitions t
      (begin
         (set-transition-tokcount t 0)
         (set-transition-predcount t (length (transition-pred t)))))
   (dolist oneplace foo
      (let ((p (find-place oneplace)))
         (initialmark p)))
   (set-graph-enabled *tpn*
      (let loop ((ts (graph-transitions *tpn*)))
         (if ts (if (enabled (car ts))
                    (cons (car ts) (loop (cdr ts)))
                    (loop (cdr ts))))))
   (set-graph-next-enabled *tpn* ())
   (set-graph-loopcount *tpn* 0))
;
;   The marking subroutines.
;
(define (elemental? p)
   (eq (place-type p) elemental))
;
(define (initialmark p)
   (set-place-token p initial)
   (dolist t (place-succ p)
      (set-transition-tokcount t (+ 1 (transition-tokcount t)))))
;
(define (mark p n)
   (set-place-token p n)
   (dolist t (place-succ p)
      (set-transition-tokcount t (+ 1 (transition-tokcount t)))))
;
(define (unmark p)
   (set-place-token p none))
;
;   The firing subroutines.
;
(define (fire t)
   (display t " fires\n")
   (let ((something (pre-fire t)))
      (dolist p (transition-pred t)
         (unmark p))
      (set-transition-tokcount t 0)
      (set-transition-firecount t (+ (transition-firecount t) 1))
      (dolist p (transition-succ t)
         (begin
            (mark p *n*)
            (dolist newt (place-succ p)
               (if (enabled newt)
                   (mark-as-enabled newt)))))
      (post-fire t something)))
;
(define (run)
   (define MODULO (length (graph-transitions *tpn*)))
   (let loop ()
      (if (= 0 (graph-constraints-left *tpn*))
         (display "All constraints satisfied . . . done.\n")
         (let ((node (get-next-to-fire)))
            (if node
               (begin
                  (fire node)
                  (loop)))))))
;
;   Now, our timed-petri-net code.
;
;   Analyzes timed petri nets.  Not memory efficient; intended to
;   show how the algorithm works.
;
(define (start)
   (define MAXN 3)
   (define *bounds* (make-array MAXN MAXN))
   (define *skew* (make-array MAXN MAXN))
   (define *preds* (make-array MAXN))
   (define *n* 0)
   (define minmax-calculated 0)
   (define skews-calculated 0)
   (define modulo 100000))
;
(start)
;
;   Because we need arrays that can get bigger at any time, we need
;   to have a routine to copy the things.
;
(define (d-make-array sizes)
   (let ((new-array (make-array (* 2 (car sizes)))))
      (if (cdr sizes)
         (forint i 0 (* 2 (car sizes))
            (aset new-array i (d-make-array (cdr sizes)))))
      new-array))
;
(define (double-array src sizes)
   (let ((new-array (make-array (* 2 (car sizes)))))
      (if (cdr sizes)
         (begin
            (forint i 0 (car sizes)
               (aset new-array i (double-array (aref src i) (cdr sizes))))
            (forint i 0 (car sizes)
               (aset new-array (+ i (car sizes))
                  (d-make-array (cdr sizes)))))
         (forint i 0 (car sizes)
            (aset new-array i (aref src i))))
      new-array))
;
(define (grow-n)
   (set! *n* (+ *n* 1))
   (if (>= *n* MAXN)
      (let ((new-MAXN (* MAXN 2)))
         (display "Growing to " new-MAXN "\n")
         (set! *bounds* (double-array *bounds* (list *n* *n*)))
         (set! *skew* (double-array *skew* (list *n* *n*)))
         (set! *preds* (double-array *preds* (list *n*)))
         (set! MAXN new-MAXN))))
;
;   The symbolic math stuff.
;
(define infty 'infty)
(define -infty '-infty)
;
(define (sym-add a b)
   (if (eq a infty)
      (if (eq b -infty)
         (error "! can't add infinity to -infinity")
         a)
      (if (eq b infty)
         (if (eq a -infty)
            (error "! can't add infinity to -infinity")
            b)
         (if (or (eq a -infty) (eq b -infty)) -infty
            (+ a b)))))
;
(define (sym-max a b)
   (if (or (eq a infty) (eq b infty)) infty
      (if (eq a -infty) b
         (if (eq b -infty) a
            (if (< a b) b a)))))
;
(define (sym-min a b)
   (if (or (eq a -infty) (eq b -infty)) -infty
      (if (eq a infty) b
         (if (eq b infty) a
            (if (< a b) a b)))))
;
(define (sym-<= a b)
   (or (eq b infty) (eq a -infty)
      (and (not (eq a infty)) (not (eq b -infty))
         (<= a b))))
;
(define (sym-neg a)
   (if (eq a infty) -infty
      (if (eq a -infty) infty
         (- 0 a))))
;
(define (sym-subset r1 r2)
   (and (sym-<= (car r1) (car r2))
        (sym-<= (car (cdr r2)) (car (cdr r1)))))
;
;   Now the code to add a node.  Strictly housekeeping.
;
(define (add-node arc-list)
   (aset *preds* *n* ())
   (dolist arc arc-list
      (let ((pred (car arc))
            (min (car (cdr arc)))
            (max (car (cdr (cdr arc)))))
         (aset *preds* *n* (cons pred (aref *preds* *n*)))
         (aset *bounds* pred *n* max)
         (aset *bounds* *n* pred (- 0 min))))
  (grow-n)
   "")
;
;   This is our skew-on-demand algorithm.  We `cache' old values.
;
(define (skew a b)
   (let ((tval (aref *skew* a b)))
      (if tval tval
         (let ((newval (recurse-skew a b)))
            (aset *skew* a b newval)
            (set! skews-calculated (+ skews-calculated 1))
            newval))))
;
;   This is the actual skew calculations.  If you do all-pairs, the
;   total complexity is (number of nodes) * (number of edges).
;   The complexity for a single pair is just the number of edges.
;
(define (minmax a b)
   (if (< a b) infty
      (if (= a b) 0
         (let* ((subval (* MODULO (/ (if (> a b) b a) MODULO)))
                (a (- a subval))
                (b (- b subval))
                (test (aref *bounds* a b)))
            (if test test
               (let ((newval (recurse-minmax a b)))
                  (aset *bounds* a b newval)
                  (set! minmax-calculated (+ minmax-calculated 1))
                  newval))))))
;
(define (recurse-minmax a b)
   (let loop ((preds (aref *preds* a)) (curmin infty))
      (if preds
         (loop (cdr preds) (sym-min curmin
                                    (sym-add (minmax (car preds) b)
                                             (aref *bounds* a (car preds)))))
         curmin)))
;
(define (recurse-skew a b)
   (sym-min (minmax a b)
      (let ((preds (aref *preds* b)))
         (if preds
             (let loop ((preds preds) (curmax -infty))
                (if preds
                   (loop (cdr preds)
                      (sym-max curmax
                         (sym-add (aref *bounds* (car preds) b)
                                  (skew a (car preds)))))
                   curmax))
             infty))))
;
;   Now we tie the two together.  Each time we fire a node, we add it to
;   our timed analysis network with the appropriate values for min and max.
;
;   We also check the timing of any constraint arcs, and report any
;   violations.
;
;   This should be done using symbolic math, but at the moment we just
;   deal with numbers.
;
;   We also need to figure out some closure conditions.
;
;   This first routine gets information before we've done the firing
;   housekeeping, conses up the information in whatever format is
;   required, and returns it.  It will be passed to post-fire.
;
(define (add-arc p t l)
   (cons (list (place-token p) (place-mintime p) (place-maxtime p)) l))
;
(define (add-c-arc p t l)
   (cons (list p (place-token p) (place-mintime p) (place-maxtime p)) l))
;
(define (pre-fire t)
   (let loop ((timing ())
              (constraints ())
              (pred-list (transition-pred t)))
      (if pred-list
         (let ((p (car pred-list)))
            (if (num? (place-token p))
               (if (elemental? p)
                  (loop (add-arc p t timing) constraints (cdr pred-list))
                  (if (place-satisfied? p)
                     (loop timing constraints (cdr pred-list))
                     (loop timing (add-c-arc p t constraints) (cdr pred-list))))
               (loop timing constraints (cdr pred-list))))
         (cons timing constraints))))
;
(define (post-fire t data)
   (let ((timing (car data))
         (constraints (cdr data))
         (this-node *n*))
      (add-node timing)
      (dolist constraint constraints
         (if (sym-subset
                (cdr (cdr constraint))
                (list (sym-neg (skew this-node (car (cdr constraint))))
                      (skew (car (cdr constraint)) this-node)))
            (begin
               (display "Constraint satisfied.\n")
               (set-graph-constraints-left *tpn*
                                 (- (graph-constraints-left *tpn*) 1))
               (set-place-satisfied? (car constraint) #t))
            (begin
               (display "***Constraint violated . . .")
               (display (cdr (cdr constraint)) " vs ")
               (display
                (list (sym-neg (skew this-node (car (cdr constraint))))
                      (skew (car (cdr constraint)) this-node)) "\n"))))
      (display "*n* " *n* " skews calculated " skews-calculated 
               " minmax calculated " minmax-calculated
               " objects active " (obactive) "\n")))
;
