;
;   Load structures.
;
(load "struct.hs")
(load "quick.hs")
;
(defstruct place
   name            ; the name of this place
   type            ; one of constraint, elemental, or spring
   mintime         ; the minimum time
   maxtime         ; the maximum time
   pred            ; predecessor transitions
   succ            ; successor transitions
   token           ; one of none, normal, or initial
   )
;
(define constraint 'constraint)
(define elemental 'elemental)
(define spring 'spring)
(define none 'none)
(define normal 'normal)
(define initial 'initial)
;
(defstruct transition
   name            ; the name of this transition
   firecount       ; how many times has it fired so far
   firingtimes     ; when did it last fire, and so on in a list
   pred            ; predecessor places; indices
   succ            ; successor transition
   tokcount        ; how many tokens in predecessor places
   predcount       ; how many predecessor places
   )
;
(defstruct graph
   name
   places         ; a list of all the places
   transitions
   enabled        ; a list of enabled transitions
   )
;
;   Allocate a global graph.
;
(define (new-tpn)
   (new)
   (define *tpn* (new-graph))
   (set-graph-name *tpn* "global-tpn"))
;
(new-tpn)
;
;   Allocate and initialize a place.
;
(define (make-place name type mintime maxtime)
   (let ((p (new-place)))
      (set-place-name p name)
      (set-place-type p type)
      (set-place-mintime p mintime)
      (set-place-maxtime p maxtime)
      (set-place-token p 'none)
      p))
;
;   Ditto, transition.
;
(define (make-transition name)
   (let ((t (new-transition)))
      (set-transition-name t name)
      (set-transition-firecount t 0)
      (set-transition-tokcount t 0)
      (set-transition-predcount t 0)
      t))
;
(define (add-transition name)
   (let ((t (make-transition name)))
      (set-graph-transitions *tpn* (cons t (graph-transitions *tpn*)))
      t))
;
(define (add-transitions . list)
   (let loop ((list list))
      (if list
         (cons
            (add-transition (car list))
            (loop (cdr list))))))
;
(define (find-transition name)
   (let loop ((list (graph-transitions *tpn*)))
      (if list
         (if (= 0 (strcmp name (transition-name (car list))))
             (car list)
             (loop (cdr list)))
         ())))
;
(define (add-place from to type mintime maxtime)
   (->q from to)
   (let ((p (make-place (strcat from "->" to) type mintime maxtime))
         (from (find-transition from))
         (to (find-transition to)))
      (set-graph-places *tpn* (cons p (graph-places *tpn*)))
      (set-place-pred p (cons from (place-pred p)))
      (set-place-succ p (cons to (place-succ p)))
      (set-transition-succ from (cons p (transition-succ from)))
      (set-transition-pred to (cons p (transition-pred to)))
      p))
;
(define (find-place name)
   (let loop ((list (graph-places *tpn*)))
      (if list
         (if (= 0 (strcmp name (place-name (car list))))
             (car list)
             (loop (cdr list)))
         ())))
;
(defmacro (for-all-places var cmd)
   `(dolist (graph-places *tpn*) ,var ,cmd))
;
(defmacro (for-all-transitions var cmd)
   `(dolist (graph-transitions *tpn*) ,var ,cmd))
;
(define (enabled t)
   (= (transition-tokcount t) (transition-predcount t)))
;
(define (mark-as-enabled t)
   (set-graph-enabled *tpn*
      (cons t (graph-enabled *tpn*))))
;
(define (get-next-to-fire)
   (let ((test (graph-enabled *tpn*)))
      (if test
         (begin
            (set-graph-enabled *tpn* (cdr test))
            (car test))
         (display "No more enabled.\n"))))
;
(define (initial-marking . foo)
   (for-all-places p
      (set-place-token p none))
   (for-all-transitions t
      (begin
         (set-transition-tokcount t 0)
         (set-transition-predcount t (length (transition-pred t)))))
   (dolist foo oneplace
      (let ((p (find-place oneplace)))
         (initialmark p)))
   (set-graph-enabled *tpn*
      (let loop ((ts (graph-transitions *tpn*)))
         (if ts (if (enabled (car ts))
                    (cons (car ts) (loop (cdr ts)))
                    (loop (cdr ts)))))))
;
;   The marking subroutines.
;
(define (initialmark p)
   (set-place-token p initial)
   (dolist (place-succ p) t
      (set-transition-tokcount t (+ 1 (transition-tokcount t)))))
;
(define (mark p)
   (set-place-token p normal)
   (dolist (place-succ p) t
      (set-transition-tokcount t (+ 1 (transition-tokcount t)))))
;
(define (unmark p)
   (set-place-token p none))
;
;   The firing subroutines.
;
(define (fire t)
   (display "Firing " t "\n")
   (dolist (transition-pred t) p
      (unmark p))
   (set-transition-tokcount t 0)
   (dolist (transition-succ t) p
      (begin
         (mark p)
         (dolist (place-succ p) newt
            (if (enabled newt)
                (mark-as-enabled newt))))))
;
(define infty 'infty)
;
(define (run)
   (let loop ()
      (fire (get-next-to-fire))
      (loop)))
